# Copyright 2024, Battelle Energy Alliance, LLC All Rights Reserved.
from montepy.data_inputs import (
    data_input,
    fill,
    importance,
    lattice_input,
    material,
    mode,
    tally,
    tally_multiplier,
    thermal_scattering,
    universe_input,
    volume,
)
from montepy.data_inputs import transform
import re

PREFIX_MATCHES = {
    fill.Fill,
    importance.Importance,
    lattice_input.LatticeInput,
    material.Material,
    mode.Mode,
    tally.Tally,
    tally_multiplier.TallyMultiplier,
    thermal_scattering.ThermalScatteringLaw,
    transform.Transform,
    volume.Volume,
    universe_input.UniverseInput,
}


def parse_data(input):
    """
    Parses the data input as the appropriate object if it is supported.

    .. versionchanged:: 0.2.0
        Removed the ``comment`` parameter, as it's in the syntax tree directly now.

    :param input: the Input object for this Data input
    :type input: Input
    :return: the parsed DataInput object
    :rtype: DataInput
    """

    base_input = data_input.DataInput(input, fast_parse=True)
    prefix = base_input.prefix

    for data_class in PREFIX_MATCHES:
        if prefix == data_class._class_prefix():
            return data_class(input)
    return data_input.DataInput(input)



# Copyright 2024, Battelle Energy Alliance, LLC All Rights Reserved.
import montepy


class CellDataPrintController:
    """
    Class for controlling if cell modifier data is printed in cell or data blocks.
    """

    def __init__(self):
        self._print_data = {}

    def __getitem__(self, key):
        if not isinstance(key, str):
            raise TypeError("Key must be a str")
        if key.upper() in montepy.Cell._ALLOWED_KEYWORDS:
            try:
                return self._print_data[key.lower()]
            except KeyError:
                return True
        else:
            raise KeyError(f"{key} is not a supported cell modifier in MCNP")

    def __setitem__(self, key, value):
        if not isinstance(key, str):
            raise TypeError("Key must be a str")
        if not isinstance(value, bool):
            raise TypeError("Must be set to a boolean value")
        if key.upper() in montepy.Cell._ALLOWED_KEYWORDS:
            self._print_data[key.lower()] = value
        else:
            raise KeyError(f"{key} is not a supported cell modifier in MCNP")

    def __str__(self):
        return f"Print data in data block: {self._print_data}"



# Copyright 2024, Battelle Energy Alliance, LLC All Rights Reserved.
from montepy.numbered_object_collection import NumberedObjectCollection
from montepy.universe import Universe


class Universes(NumberedObjectCollection):
    """
    A container of multiple :class:`~montepy.universe.Universe` instances.
    """

    def __init__(self, objects=None, problem=None):
        super().__init__(Universe, objects, problem)



# Copyright 2024, Battelle Energy Alliance, LLC All Rights Reserved.
import montepy
from montepy.numbered_object_collection import NumberedObjectCollection

Material = montepy.data_inputs.material.Material


class Materials(NumberedObjectCollection):
    """
    A container of multiple :class:`~montepy.data_inputs.material.Material` instances.

    :param objects: the list of materials to start with if needed
    :type objects: list
    """

    def __init__(self, objects=None, problem=None):
        super().__init__(Material, objects, problem)



# Copyright 2024, Battelle Energy Alliance, LLC All Rights Reserved.
import copy
import math
import montepy
from montepy.errors import *
import os

from unittest import TestCase


class EdgeCaseTests(TestCase):
    def test_complement_edge_case(self):
        capsule = montepy.read_input("tests/inputs/test_complement_edge.imcnp")
        complementing_cell = capsule.cells[61482]
        self.assertEqual(len(complementing_cell.complements), 2)

    def test_surface_edge_case(self):
        capsule = montepy.read_input("tests/inputs/test_complement_edge.imcnp")
        problem_cell = capsule.cells[61441]
        self.assertEqual(len(set(problem_cell.surfaces)), 6)

    def test_interp_surface_edge_case(self):
        capsule = montepy.read_input("tests/inputs/test_interp_edge.imcnp")
        self.assertEqual(len(capsule.cells[10214].surfaces), 4)

    def test_excess_mt(self):
        with self.assertRaises(MalformedInputError):
            montepy.read_input(os.path.join("tests", "inputs", "test_excess_mt.imcnp"))

    def test_missing_mat_for_mt(self):
        with self.assertRaises(MalformedInputError):
            montepy.read_input(
                os.path.join("tests", "inputs", "test_missing_mat_for_mt.imcnp")
            )

    def test_orphaning_mt(self):
        problem = montepy.read_input(os.path.join("tests", "inputs", "test.imcnp"))
        card = montepy.input_parser.mcnp_input.Input(
            ["MT5 lwtr.01t"],
            montepy.input_parser.block_type.BlockType.DATA,
        )
        problem.data_inputs.append(montepy.data_inputs.data_parser.parse_data(card))
        try:
            with self.assertRaises(MalformedInputError):
                problem.write_to_file("out")
        finally:
            if os.path.exists("out"):
                pass
                os.remove("out")

    def test_shortcuts_in_special_comment(self):
        in_str = "fc247 experiment in I24 Cell Specific Heating"
        card = montepy.input_parser.mcnp_input.Input(
            [in_str], montepy.input_parser.block_type.BlockType.DATA
        )
        self.assertEqual(card.input_lines, [in_str])
        in_str = in_str.replace("fc247", "sc247")
        card = montepy.input_parser.mcnp_input.Input(
            [in_str], montepy.input_parser.block_type.BlockType.DATA
        )
        self.assertEqual(card.input_lines, [in_str])

    def test_long_lines(self):
        with self.assertWarns(montepy.errors.LineOverRunWarning):
            problem = montepy.read_input(
                "tests/inputs/test_long_lines.imcnp", (5, 1, 60)
            )
            comment = problem.cells[1].comments[0]
            self.assertTrue(len(comment.contents[0]) <= 80)
            self.assertEqual(len(problem.surfaces), 3)

    def test_confused_key_word(self):
        # this came from Andrew Bascom's Issue #99
        in_strs = [
            "29502 30002 2.53E-02 ( 500020 -500023  -500060 ):  $ Outer Capsule Lower Endcap",
            "          ( 500023 -500024 500062 -500060 ):  $ Outer Capsule Lower Endcap",
            "          ( 500024 -500025 500062 -500061 )  $ Outer Capsule Lower Endcap",
            "                u= 106 $ Outer Capsule Lower Endcap",
        ]
        input = montepy.input_parser.mcnp_input.Input(
            in_strs, montepy.input_parser.block_type.BlockType.CELL
        )
        cell = montepy.Cell(input)
        self.assertNotIn("", cell.parameters)
        print(cell.parameters)
        allowed_keys = {"u", "imp:n", "fill", "lat", "vol"}
        self.assertEqual(len(cell.parameters), 5)
        self.assertEqual(cell.parameters.nodes.keys(), allowed_keys)

    def test_confused_union_geometry(self):
        # based on issue #122
        in_strs = [
            "9800     10    -0.123000 +101 -200 -905 +213 (-216:+217)",
        ]
        input = montepy.input_parser.mcnp_input.Input(
            in_strs, montepy.input_parser.block_type.BlockType.CELL
        )
        cell = montepy.Cell(input)

    def test_confused_trcl(self):
        for line in [
            "340 0 (94 -209 -340) fill=2 trcl=(0 0 0  0.7071 -0.7071 0  0.7071 0.7071 0  0 0 1)",
            "340 0 (94 -209 -340) fill=2 trcl=(0 0 0  0.7071 -0.7071 0  0.7071 0.7071 0  0 0 1)  $ Comment",
        ]:
            input = montepy.input_parser.mcnp_input.Input(
                [line],
                montepy.input_parser.block_type.BlockType.CELL,
            )
            cell = montepy.Cell(input)

    def test_space_after_equals(self):
        lines = [
            "35000 3690   0.01000    35100   -35105",
            "                        35110   -35115",
            "                        35150   -35155   U = 35100",
        ]
        input = montepy.input_parser.mcnp_input.Input(
            lines,
            montepy.input_parser.block_type.BlockType.CELL,
        )
        cell = montepy.Cell(input)

    def test_interpolate_geometry(self):
        lines = [
            "10234   30  1.2456780      -3103  3104 -3133  3136",
            "                            3201  15i   3217",
            "                            u=20",
        ]
        input = montepy.input_parser.mcnp_input.Input(
            lines,
            montepy.input_parser.block_type.BlockType.CELL,
        )
        cell = montepy.Cell(input)

    def test_material_float(self):
        in_strs = ["m171 1001.80c 1E-24"]
        input = montepy.input_parser.mcnp_input.Input(
            in_strs, montepy.input_parser.block_type.BlockType.CELL
        )
        material = montepy.data_inputs.material.Material(input)

    def test_void_cell_set_material(self):
        in_strs = ["2 0 -63 -62 64  imp:n=1 $ COBALT_TARGET_PELLET2"]
        input = montepy.input_parser.mcnp_input.Input(
            in_strs, montepy.input_parser.block_type.BlockType.CELL
        )
        base_cell = montepy.Cell(input)
        in_strs = ["m171 1001.80c 1E-24"]
        input = montepy.input_parser.mcnp_input.Input(
            in_strs, montepy.input_parser.block_type.BlockType.CELL
        )
        material = montepy.data_inputs.material.Material(input)
        base_cell.material = material
        for dens_value in {5.0, 1.235e-6, 1.245e23}:
            cell = copy.deepcopy(base_cell)
            print(dens_value)
            cell.mass_density = dens_value
            with self.assertWarns(montepy.errors.LineExpansionWarning):
                output = cell.format_for_mcnp_input((6, 2, 0))
            print(output)
            parts = output[0].split()
            self.assertAlmostEqual(float(parts[2]), -dens_value, places=9)

    def test_geometry_comments(self):
        in_strs = """21073   130 0.010000   (-11516    97  -401 )  $ C 1 Lower water
      :(-11526    97  -401 )  $ C 2 Lower water
      :(-11536    97  -401 )  $ C 3 Lower water
      :(-11546    97  -401 )  $ C 4 Lower water
      :(-11556    97  -401 )  $ C 5 Lower water
      :(-11576    97  -401 ) imp:n=1 $ C 7 Lower water""".split(
            "\n"
        )
        input = montepy.input_parser.mcnp_input.Input(
            in_strs, montepy.input_parser.block_type.BlockType.CELL
        )
        cell = montepy.Cell(input)
        # this step caused an error for #163
        cell.comments
        cell._tree.format()



# Copyright 2024, Battelle Energy Alliance, LLC All Rights Reserved.
from unittest import TestCase

import montepy
from montepy.geometry_operators import Operator
from montepy.input_parser import syntax_node
from montepy.surfaces.half_space import HalfSpace, UnitHalfSpace


class TestHalfSpaceUnit(TestCase):
    def test_init(self):
        surface = montepy.surfaces.CylinderOnAxis()
        node = montepy.input_parser.syntax_node.GeometryTree("hi", {}, "*", " ", " ")
        half_space = HalfSpace(+surface, Operator.UNION, -surface, node)
        self.assertIs(half_space.operator, Operator.UNION)
        self.assertEqual(half_space.left, +surface)
        self.assertEqual(half_space.right, -surface)
        self.assertIs(half_space.node, node)
        with self.assertRaises(TypeError):
            HalfSpace(surface, Operator.UNION)
        with self.assertRaises(TypeError):
            HalfSpace(+surface, "hi")
        with self.assertRaises(TypeError):
            HalfSpace(+surface, Operator.UNION, surface)
        with self.assertRaises(TypeError):
            HalfSpace(+surface, Operator.UNION, -surface, "hi")
        with self.assertRaises(ValueError):
            HalfSpace(+surface, Operator.UNION)
        with self.assertRaises(ValueError):
            HalfSpace(+surface, Operator.INTERSECTION)

    def test_get_leaves(self):
        surface = montepy.surfaces.CylinderOnAxis()
        cell = montepy.Cell()
        half_space = -surface & ~cell
        cells, surfaces = half_space._get_leaf_objects()
        self.assertEqual(cells, {cell})
        self.assertEqual(surfaces, {surface})

    def test_len(self):
        surface = montepy.surfaces.CylinderOnAxis()
        cell = montepy.Cell()
        half_space = -surface & ~cell
        self.assertEqual(len(half_space), 2)

    def test_eq(self):
        cell1 = montepy.Cell()
        cell2 = montepy.Cell()
        half1 = ~cell1 & ~cell2
        self.assertTrue(half1 == half1)
        half2 = ~cell1 | ~cell2
        self.assertTrue(half1 != half2)
        half2 = ~cell1 & ~cell1
        self.assertTrue(half1 != half2)
        half2 = ~cell2 & ~cell2
        self.assertTrue(half1 != half2)
        half2 = ~half1
        half2._operator = Operator.INTERSECTION
        self.assertTrue(half1 != half2)
        with self.assertRaises(TypeError):
            half1 == "hi"

    def test_str(self):
        cell1 = montepy.Cell()
        cell1.number = 1
        cell2 = montepy.Cell()
        cell2.number = 2
        half_space = ~cell1 | ~cell2
        self.assertEqual(str(half_space), "(#1:#2)")
        half_space = ~cell1 & ~cell2
        self.assertEqual(str(half_space), "(#1*#2)")
        half_space = ~(~cell1 & ~cell2)
        self.assertEqual(str(half_space), "#(#1*#2)")
        # test that no errors occur
        repr(half_space)


class TestUnitHalfSpaceUnit(TestCase):
    def test_init(self):
        node = montepy.input_parser.syntax_node.ValueNode("123", float)
        half_space = UnitHalfSpace(123, True, False, node)
        self.assertEqual(half_space.divider, 123)
        self.assertTrue(half_space.side)
        self.assertTrue(not half_space.is_cell)
        self.assertIs(half_space.node, node)
        with self.assertRaises(TypeError):
            UnitHalfSpace("hi", True, False)
        with self.assertRaises(TypeError):
            half_space = UnitHalfSpace(123, "hi", False)
        with self.assertRaises(TypeError):
            half_space = UnitHalfSpace(123, True, "hi")
        with self.assertRaises(TypeError):
            half_space = UnitHalfSpace(123, True, False, "hi")

    def test_eq(self):
        half1 = UnitHalfSpace(1, True, False)
        self.assertTrue(half1 == half1)
        half2 = UnitHalfSpace(2, True, False)
        self.assertTrue(half1 != half2)
        half2 = UnitHalfSpace(1, False, False)
        self.assertTrue(half1 != half2)
        half2 = UnitHalfSpace(1, True, True)
        self.assertTrue(half1 != half2)
        with self.assertRaises(TypeError):
            half1 == "hi"

    def test_str(self):
        half_space = UnitHalfSpace(1, True, False)
        self.assertEqual(str(half_space), "+1")
        half_space.side = False
        self.assertEqual(str(half_space), "-1")
        half_space.is_cell = True
        self.assertEqual(str(half_space), "1")
        cell = montepy.Cell()
        cell.number = 1
        half_space.divider = cell
        self.assertEqual(str(half_space), "1")


class TestGeometryIntegration(TestCase):
    def test_surface_half_space(self):
        surface = montepy.surfaces.cylinder_on_axis.CylinderOnAxis()
        half_space = +surface
        self.assertIsInstance(half_space, HalfSpace)
        self.assertIsInstance(half_space, UnitHalfSpace)
        self.assertIs(half_space.divider, surface)
        self.assertTrue(half_space.side)
        half_space = -surface
        self.assertIsInstance(half_space, HalfSpace)
        self.assertIsInstance(half_space, UnitHalfSpace)
        self.assertIs(half_space.divider, surface)
        self.assertTrue(not half_space.side)
        with self.assertRaises(TypeError):
            half_space = surface & surface
        with self.assertRaises(TypeError):
            half_space = surface | surface
        with self.assertRaises(TypeError):
            half_space = ~surface

    def test_cell_half_space(self):
        cell = montepy.Cell()
        half_space = ~cell
        self.assertIsInstance(half_space, HalfSpace)
        self.assertIs(half_space.left.divider, cell)
        self.assertTrue(half_space.left.side)
        with self.assertRaises(TypeError):
            half_space = cell & cell
        with self.assertRaises(TypeError):
            half_space = cell | cell
        with self.assertRaises(TypeError):
            half_space = +cell
        with self.assertRaises(TypeError):
            half_space = -cell

    def test_intersect_half_space(self):
        cell1 = ~montepy.Cell()
        cell2 = ~montepy.Cell()
        half_space = cell1 & cell2
        self.assertIsInstance(half_space, HalfSpace)
        self.assertIs(half_space.operator, Operator.INTERSECTION)
        self.assertIs(half_space.left, cell1)
        self.assertIs(half_space.right, cell2)
        with self.assertRaises(TypeError):
            cell1 & "hi"
        with self.assertRaises(TypeError):
            "hi" & cell1
        with self.assertRaises(TypeError):
            cell2 + cell1
        with self.assertRaises(TypeError):
            cell2 - cell1

    def test_union_half_space(self):
        cell1 = ~montepy.Cell()
        cell2 = ~montepy.Cell()
        half_space = cell1 | cell2
        self.assertIsInstance(half_space, HalfSpace)
        self.assertIs(half_space.operator, Operator.UNION)
        self.assertIs(half_space.left, cell1)
        self.assertIs(half_space.right, cell2)
        with self.assertRaises(TypeError):
            cell1 | "hi"
        with self.assertRaises(TypeError):
            "hi" | cell1

    def test_invert_half_space(self):
        cell1 = ~montepy.Cell()
        cell2 = ~montepy.Cell()
        half_space1 = cell1 | cell2
        half_space = ~half_space1
        self.assertIsInstance(half_space, HalfSpace)
        self.assertIs(half_space.operator, Operator.COMPLEMENT)
        self.assertIs(half_space.left, half_space1)
        self.assertIsNone(half_space.right)

    def test_iand_recursion(self):
        cell1 = montepy.Cell()
        cell2 = montepy.Cell()
        cell3 = montepy.Cell()
        half_space = ~cell1 & ~cell2
        cell1.number = 1
        cell2.number = 2
        cell3.number = 3
        cell3.geometry = half_space
        half_space &= ~cell1
        self.assertEqual(half_space.left, ~cell1)
        self.assertNotIsInstance(half_space.right, UnitHalfSpace)
        self.assertEqual(half_space.right.left, ~cell2)
        self.assertEqual(half_space.right.right, ~cell1)
        self.assertEqual(half_space.operator, Operator.INTERSECTION)
        self.assertEqual(len(half_space), 3)
        for i in range(1, 100):
            half_space &= ~cell1
            self.assertEqual(len(half_space), i + 3)
        with self.assertRaises(TypeError):
            half_space &= "hi"
        # test with unit halfspaces
        surf = montepy.surfaces.CylinderParAxis()
        # test going from leaf to tree
        half_space = -surf
        half_space &= +surf
        self.assertEqual(len(half_space), 2)
        # test with actual tree
        half_space = -surf | +surf
        half_space &= +surf
        self.assertEqual(len(half_space), 3)

    def test_ior_recursion(self):
        cell1 = ~montepy.Cell()
        cell2 = ~montepy.Cell()
        half_space = cell1 | cell2
        half_space |= cell1
        self.assertIs(half_space.left, cell1)
        self.assertNotIsInstance(half_space.right, UnitHalfSpace)
        self.assertEqual(half_space.right.left, cell2)
        self.assertEqual(half_space.right.right, cell1)
        self.assertEqual(half_space.operator, Operator.UNION)
        self.assertEqual(len(half_space), 3)
        for i in range(1, 100):
            half_space |= cell1
            self.assertEqual(len(half_space), i + 3)
        with self.assertRaises(TypeError):
            half_space |= "hi"
        # test with unit halfspaces
        surf = montepy.surfaces.CylinderParAxis()
        half_space = -surf
        half_space |= +surf
        self.assertEqual(len(half_space), 2)
        half_space = -surf | +surf
        half_space |= +surf
        self.assertEqual(len(half_space), 3)

    def test_parse_input_tree(self):
        # simplest intersection
        input = montepy.input_parser.mcnp_input.Input(
            ["1 0 1 -2"], montepy.input_parser.block_type.BlockType.CELL
        )
        parser = montepy.input_parser.cell_parser.CellParser()
        tree = parser.parse(input.tokenize())
        geometry = tree["geometry"]
        half_space = HalfSpace.parse_input_node(geometry)
        self.assertEqual(half_space.operator, Operator.INTERSECTION)
        self.assertEqual(half_space.left.divider, 1)
        self.assertTrue(half_space.left.side)
        self.assertTrue(not half_space.left.is_cell)
        self.assertEqual(half_space.right.divider, 2)
        self.assertTrue(not half_space.right.side)
        self.assertTrue(not half_space.right.is_cell)
        # Test type checking
        with self.assertRaises(TypeError):
            HalfSpace.parse_input_node("hi")
        # test assymetric
        input = montepy.input_parser.mcnp_input.Input(
            ["1 0 #2"], montepy.input_parser.block_type.BlockType.CELL
        )
        parser = montepy.input_parser.cell_parser.CellParser()
        tree = parser.parse(input.tokenize())
        geometry = tree["geometry"]
        half_space = HalfSpace.parse_input_node(geometry)
        self.assertEqual(half_space.operator, Operator.COMPLEMENT)
        self.assertEqual(half_space.left.divider, 2)
        self.assertTrue(half_space.left.side)
        self.assertTrue(half_space.left.is_cell)
        self.assertIsNone(half_space.right)
        # Test nested trees
        input = montepy.input_parser.mcnp_input.Input(
            ["1 0 1 -2 : 3"], montepy.input_parser.block_type.BlockType.CELL
        )
        parser = montepy.input_parser.cell_parser.CellParser()
        tree = parser.parse(input.tokenize())
        geometry = tree["geometry"]
        half_space = HalfSpace.parse_input_node(geometry)
        self.assertEqual(half_space.operator, Operator.UNION)
        self.assertNotIsInstance(half_space.left, UnitHalfSpace)
        self.assertIsInstance(half_space.right, UnitHalfSpace)
        self.assertEqual(half_space.left.operator, Operator.INTERSECTION)
        self.assertEqual(half_space.right.divider, 3)
        self.assertEqual(half_space.left.left.divider, 1)
        self.assertEqual(half_space.left.right.divider, 2)
        # test shift
        input = montepy.input_parser.mcnp_input.Input(
            ["1 0 #(1 2 3)"], montepy.input_parser.block_type.BlockType.CELL
        )
        parser = montepy.input_parser.cell_parser.CellParser()
        tree = parser.parse(input.tokenize())
        geometry = tree["geometry"]
        half_space = HalfSpace.parse_input_node(geometry)
        self.assertEqual(half_space.operator, Operator.COMPLEMENT)
        self.assertIsNone(half_space.right)
        self.assertEqual(half_space.left.operator, Operator.INTERSECTION)

    def test_update_tree_with_comment(self):
        input = montepy.input_parser.mcnp_input.Input(
            ["1 0 -1 $hi there", "      2"],
            montepy.input_parser.block_type.BlockType.CELL,
        )
        cell = montepy.Cell(input)
        geometry = cell.geometry
        geometry._update_values()

    def test_parse_input_value_node(self):
        node = montepy.input_parser.syntax_node.ValueNode("-1", float)
        half_space = UnitHalfSpace.parse_input_node(node)
        self.assertTrue(not half_space.is_cell)
        self.assertTrue(not half_space.side)
        self.assertEqual(half_space.divider, 1)
        half_space = UnitHalfSpace.parse_input_node(node, True)
        self.assertTrue(half_space.is_cell)
        self.assertTrue(half_space.side)
        self.assertEqual(half_space.divider, 1)
        node = montepy.input_parser.syntax_node.ValueNode("+1", float)
        half_space = UnitHalfSpace.parse_input_node(node, False)
        self.assertTrue(not half_space.is_cell)
        self.assertTrue(half_space.side)
        self.assertEqual(half_space.divider, 1)
        with self.assertRaises(TypeError):
            UnitHalfSpace.parse_input_node("hi", False)
        with self.assertRaises(TypeError):
            UnitHalfSpace.parse_input_node(node, "hi")

    def test_unit_divider_setter(self):
        cell = montepy.Cell()
        cell.number = 1
        parent = montepy.Cell()
        parent.number = 2
        node = montepy.input_parser.syntax_node.ValueNode("1", float)
        half_space = UnitHalfSpace.parse_input_node(node, True)
        cells = montepy.cells.Cells()
        cells.append(cell)
        half_space.update_pointers(cells, [], parent)
        self.assertIs(half_space.divider, cell)
        cell2 = montepy.Cell()
        # #madLads
        cell2.number = 4
        half_space.divider = cell2
        self.assertIs(half_space.divider, cell2)
        self.assertIn(cell2, parent.complements)
        # test with surface
        surf = montepy.surfaces.CylinderParAxis()
        surf.number = 5
        with self.assertRaises(TypeError):
            half_space.divider = surf
        half_space.is_cell = False
        half_space.divider = surf
        self.assertIs(half_space.divider, surf)
        self.assertIn(surf, parent.surfaces)
        with self.assertRaises(TypeError):
            half_space.divider = "hi"

    def test_unit_create_default_node(self):
        surf = montepy.surfaces.CylinderParAxis()
        surf.number = 1
        half_space = UnitHalfSpace(surf, True, False)
        half_space._ensure_has_nodes()
        node = half_space.node
        self.assertIsInstance(node, syntax_node.ValueNode)
        self.assertEqual(node.value, 1)
        self.assertEqual(node.format(), "1")
        # test with negative side
        half_space = UnitHalfSpace(surf, False, False)
        half_space._ensure_has_nodes()
        node = half_space.node
        self.assertEqual(node.format(), "-1")
        # test with int and not an object
        half_space = UnitHalfSpace(1, True, False)
        half_space._ensure_has_nodes()
        node = half_space.node
        self.assertIsInstance(node, syntax_node.ValueNode)
        self.assertEqual(node.value, 1)

    def test_tree_create_default_node(self):
        surf = montepy.surfaces.CylinderParAxis()
        surf.number = 1
        half_space = -surf | +surf
        half_space._ensure_has_nodes()
        node = half_space.node
        self.assertIsInstance(node, syntax_node.GeometryTree)
        self.assertEqual(node.operator, Operator.UNION)
        self.assertEqual(node.format(), "-1 : 1")
        # Test intersection
        half_space = -surf & +surf
        half_space._ensure_has_nodes()
        node = half_space.node
        self.assertEqual(node.operator, Operator.INTERSECTION)
        # Test complement
        half_space = ~(-surf | +surf)
        half_space._ensure_has_nodes()
        node = half_space.node
        self.assertEqual(node.operator, Operator.COMPLEMENT)
        self.assertEqual(node.format(), " #(-1 : 1)")
        # test cell complement
        cell = montepy.Cell()
        cell.number = 1
        half_space = ~cell
        half_space._ensure_has_nodes()
        node = half_space.node
        self.assertEqual(node.operator, Operator.COMPLEMENT)
        self.assertEqual(node.format(), " #1")

    def test_update_operators_in_node(self):
        surf = montepy.surfaces.CylinderParAxis()
        surf.number = 1
        half_space = -surf | +surf
        half_space._ensure_has_nodes()
        half_space.operator = Operator.INTERSECTION
        half_space._update_values()
        self.assertEqual(half_space.node.format(), "-1   1")
        del half_space.right
        half_space.operator = Operator.COMPLEMENT
        cell = montepy.Cell()
        cell.number = 1
        half_space.left = UnitHalfSpace(cell, True, True)
        half_space._update_values()
        self.assertEqual(half_space.node.format(), "  #1")
        # test with blank tree
        half_space = -surf & +surf
        half_space._update_values()
        self.assertEqual(half_space.node.format(), "-1 1")
        # test move to union
        # test centering
        half_space.node.nodes["operator"].append("  ")
        half_space.operator = Operator.UNION
        half_space._update_values()
        self.assertEqual(half_space.node.format(), "-1 : 1")



import glob
import os
import sys
import warnings

ignored = {"__pycache__", "_version.py", "__main__.py", "_cell_data_control.py"}

base = os.path.join("..", "..")


def crawl_path(rel_path):
    missing = False
    for f in os.listdir(os.path.join(base, rel_path)):
        f_name = os.path.join(rel_path, f)
        if f in ignored:
            continue
        if os.path.isdir(os.path.join(base, f_name)):
            crawl_path(f_name)
        elif os.path.isfile(os.path.join(base, f_name)) and ".py" in f:
            if f == "__init__.py":
                path = f_name.replace("/", ".").replace(".__init__.py", ".rst")
            else:
                path = f_name.replace("/", ".").replace(".py", ".rst")
            if not os.path.exists(os.path.join("api", path)):
                missing = True
                warnings.warn(
                    f"Missing sphinx documentation for {os.path.join(rel_path, f)}"
                )
    return missing


missing = crawl_path("montepy")
if missing:
    sys.exit(314)



# Copyright 2024, Battelle Energy Alliance, LLC All Rights Reserved.
import itertools
from montepy.data_inputs.cell_modifier import CellModifierInput
from montepy.data_inputs.lattice import Lattice
from montepy.errors import *
from montepy.input_parser.mcnp_input import Jump
from montepy.input_parser import syntax_node
from montepy.mcnp_object import MCNP_Object
from montepy.utilities import *


class LatticeInput(CellModifierInput):
    """
    Object to handle the inputs from ``LAT``.

    :param input: the Input object representing this data input
    :type input: Input
    :param in_cell_block: if this card came from the cell block of an input file.
    :type in_cell_block: bool
    :param key: the key from the key-value pair in a cell
    :type key: str
    :param value: the value syntax tree from the key-value pair in a cell
    :type value: SyntaxNode
    """

    def __init__(self, input=None, in_cell_block=False, key=None, value=None):
        super().__init__(input, in_cell_block, key, value)
        self._lattice = self._generate_default_node(int, None)
        self._lattice._convert_to_enum(Lattice, True, int)
        if self.in_cell_block:
            if key:
                try:
                    val = value["data"][0]
                    val._convert_to_int()
                    val._convert_to_enum(Lattice, int)
                except ValueError as e:
                    raise ValueError("Cell Lattice must be 1 or 2")
                self._lattice = val
        elif input:
            self._lattice = []
            words = self.data
            for word in words:
                try:
                    word._convert_to_int()
                    word._convert_to_enum(Lattice, int)
                    self._lattice.append(word)
                except ValueError:
                    raise MalformedInputError(
                        input, f"Cell lattice must be 1 or 2. {word} given."
                    )

    def _generate_default_cell_tree(self):
        list_node = syntax_node.ListNode("number sequence")
        data = self._generate_default_node(int, None)
        data._convert_to_enum(Lattice, True, int)
        list_node.append(data)
        classifier = syntax_node.ClassifierNode()
        classifier.prefix = self._generate_default_node(
            str, self._class_prefix().upper(), None
        )
        self._tree = syntax_node.SyntaxNode(
            "lattice",
            {
                "classifier": classifier,
                "param_seperator": self._generate_default_node(str, "=", None),
                "data": list_node,
            },
        )

    @staticmethod
    def _class_prefix():
        return "lat"

    @staticmethod
    def _has_number():
        return False

    @staticmethod
    def _has_classifier():
        return 0

    @property
    def has_information(self):
        if self.in_cell_block:
            return self.lattice is not None

    @make_prop_val_node("_lattice", (Lattice, int, type(None)), Lattice, deletable=True)
    def lattice(self):
        """
        The type of lattice being used.

        :rtype: Lattice
        """
        pass

    @property
    def _tree_value(self):
        return self._lattice

    def push_to_cells(self):
        if self._problem and not self.in_cell_block:
            cells = self._problem.cells
            if self._lattice:
                self._check_redundant_definitions()
                for cell, lattice in itertools.zip_longest(
                    cells, self._lattice, fillvalue=None
                ):
                    if not isinstance(lattice, (Jump, type(None))):
                        cell._lattice._lattice = lattice

    def merge(self, other):
        raise MalformedInputError(
            other._input, "Cannot have two lattice inputs for the problem"
        )

    def _clear_data(self):
        del self._lattice

    def __str__(self):
        return "Lattice: {self.lattice}"

    def __repr__(self):
        ret = (
            f"Lattice: in_cell: {self._in_cell_block}"
            f" set_in_block: {self.set_in_cell_block}, "
            f"Lattice_values : {self.lattice}"
        )
        return ret

    def _update_cell_values(self):
        pass



# Copyright 2024, Battelle Energy Alliance, LLC All Rights Reserved.
from unittest import TestCase

import montepy
from montepy.data_inputs.data_parser import parse_data
from montepy.input_parser.block_type import BlockType


class TestSourceDefinition(TestCase):
    def test_complex_sdef_parser(self):
        lines = ["si1  l   60001 838i 60840", " sp1  d 0.0100 0.01000 0.0100"]
        for test_line in lines:
            print(test_line)
            input = montepy.input_parser.mcnp_input.Input([test_line], BlockType.DATA)
            parse_data(input)



# Copyright 2024, Battelle Energy Alliance, LLC All Rights Reserved.
from unittest import TestCase

import montepy
from montepy.cell import Cell
from montepy.input_parser.block_type import BlockType
from montepy.input_parser.mcnp_input import Input


class TestCellClass(TestCase):
    def test_bad_init(self):
        with self.assertRaises(TypeError):
            Cell("5")

    def test_init(self):
        # test invalid cell number
        in_str = "foo"
        card = Input([in_str, f"c {in_str}"], BlockType.CELL)
        with self.assertRaises(montepy.errors.MalformedInputError):
            in_str = "foo bar"
            cell = Cell(card)
        # test like feature unsupported
        in_str = "1 like 2"
        card = Input([in_str, f"c {in_str}"], BlockType.CELL)
        with self.assertRaises(montepy.errors.UnsupportedFeature):
            in_str = "foo bar"
            cell = Cell(card)
        # test invalid material number
        in_str = "1 foo"
        card = Input([in_str], BlockType.CELL)
        with self.assertRaises(montepy.errors.MalformedInputError):
            cell = Cell(card)
        # test invalid material density
        in_str = "1 1 foo"
        card = Input([in_str], BlockType.CELL)
        with self.assertRaises(montepy.errors.MalformedInputError):
            cell = Cell(card)
        # tests void cell
        in_str = "1 0 2"
        card = Input([in_str], BlockType.CELL)
        cell = Cell(card)
        self.assertEqual(cell.old_number, 1)
        self.assertEqual(cell.number, 1)
        self.assertIsNone(cell.material)
        self.assertEqual(cell.old_mat_number, 0)

        # test material cell
        for atom_dens, density in [(False, "-0.5"), (True, "0.5")]:
            in_str = f"1 1 {density} 2"
            card = Input([in_str], BlockType.CELL)
            cell = Cell(card)
            self.assertEqual(cell.old_mat_number, 1)
            if atom_dens:
                self.assertAlmostEqual(cell.atom_density, 0.5)
            else:
                self.assertAlmostEqual(cell.mass_density, 0.5)
            self.assertTrue(atom_dens == cell.is_atom_dens)

        # test parameter input
        in_str = "1 0 #2 u= 5 vol=20 trcl=5"
        card = Input([in_str], BlockType.CELL)
        cell = Cell(card)
        self.assertEqual(cell.parameters["TRCL"]["data"][0].value, 5.0)

    # TODO test updating cell geometry once done
    def test_cell_validator(self):
        cell = Cell()
        with self.assertRaises(montepy.errors.IllegalState):
            cell.validate()
        with self.assertRaises(montepy.errors.IllegalState):
            cell.format_for_mcnp_input((6, 2, 0))
        cell.mass_density = 5.0
        with self.assertRaises(montepy.errors.IllegalState):
            cell.validate()
        del cell.mass_density

    # TODO test geometry stuff

    def test_number_setter(self):
        in_str = "1 0 2"
        card = Input([in_str], BlockType.CELL)
        cell = Cell(card)
        cell.number = 5
        self.assertEqual(cell.number, 5)
        with self.assertRaises(TypeError):
            cell.number = "5"
        with self.assertRaises(ValueError):
            cell.number = -5

    def test_cell_density_setter(self):
        in_str = "1 1 0.5 2"
        card = Input([in_str], BlockType.CELL)
        cell = Cell(card)
        cell.mass_density = 1.5
        self.assertEqual(cell._density, 1.5)
        self.assertEqual(cell.mass_density, 1.5)
        self.assertFalse(cell.is_atom_dens)
        with self.assertRaises(AttributeError):
            _ = cell.atom_density
        cell.atom_density = 1.6
        self.assertEqual(cell._density, 1.6)
        self.assertEqual(cell.atom_density, 1.6)
        self.assertTrue(cell.is_atom_dens)
        with self.assertRaises(AttributeError):
            _ = cell.mass_density
        with self.assertRaises(TypeError):
            cell.atom_density = (5, True)
        with self.assertRaises(TypeError):
            cell.mass_density = "five"
        with self.assertRaises(ValueError):
            cell.atom_density = -1.5
        with self.assertRaises(ValueError):
            cell.mass_density = -5

    def test_cell_density_deleter(self):
        in_str = "1 1 0.5 2"
        card = Input([in_str], BlockType.CELL)
        cell = Cell(card)
        del cell.mass_density
        self.assertIsNone(cell.mass_density)
        cell.atom_density = 1.0
        del cell.atom_density
        self.assertIsNone(cell.atom_density)

    def test_cell_sorting(self):
        in_str = "1 1 0.5 2"
        card = Input([in_str], BlockType.CELL)
        cell1 = Cell(card)
        in_str = "2 1 0.5 2"
        card = Input([in_str], BlockType.CELL)
        cell2 = Cell(card)
        test_sort = sorted([cell2, cell1])
        answer = [cell1, cell2]
        for i, cell in enumerate(test_sort):
            self.assertEqual(cell, answer[i])

    def test_cell_parameters_setting(self):
        in_str = "1 1 0.5 2"
        card = Input([in_str], BlockType.CELL)
        cell = Cell(card)
        params = {"FILL": "5"}
        cell.parameters = params
        self.assertEqual(params, cell.parameters)
        with self.assertRaises(TypeError):
            cell.parameters = []

    def test_cell_str(self):
        in_str = "1 1 0.5 2"
        card = Input([in_str], BlockType.CELL)
        cell = Cell(card)
        self.assertEqual(str(cell), "CELL: 1, mat: 0, DENS: 0.5 atom/b-cm")
        self.assertEqual(
            repr(cell), "CELL: 1 \nVoid material \ndensity: 0.5 atom/b-cm\n\n"
        )

    def test_cell_paremeters_no_eq(self):
        in_str = f"1 0 -1 PWT 1.0"
        card = Input([in_str], BlockType.CELL)
        cell = Cell(card)
        self.assertEqual(cell.parameters["PWT"]["data"][0].value, 1.0)



# Copyright 2024, Battelle Energy Alliance, LLC All Rights Reserved.
from enum import Enum


class Operator(Enum):
    """
    Enumeration of the allowed geometry set logic.
    """

    INTERSECTION = "*"
    """
    Represents the intersection of sets.
    """
    UNION = ":"
    """
    Represents the union of sets.
    """
    COMPLEMENT = "#"
    """
    Represents the complement of a set.
    """
    _SHIFT = ">"
    """
    Internal operator essentially equivalent to No-op.

    This is used to properly handle parentheses while parsing.
    """



# Copyright 2024, Battelle Energy Alliance, LLC All Rights Reserved.
import os
import unittest

import montepy
from montepy.input_parser.input_file import MCNP_InputFile


class testInputFileWrapper(unittest.TestCase):
    """
    Note: file lineno is 1-indexed as is tradition in file editors.
    """

    def setUp(self):
        self._file_name = os.path.join("tests", "inputs", "test.imcnp")
        self._file = MCNP_InputFile(self._file_name)

    def test_init(self):
        self.assertEqual(self._file.name, self._file_name)
        parent = os.path.join("tests", "inputs", "testRead.imcnp")
        test_file = MCNP_InputFile(
            os.path.join("tests", "inputs", "testReadTarget.imcnp"), parent
        )
        self.assertEqual(test_file.parent_file, parent)

    def test_open_iter(self):
        with self._file.open("r") as fh:
            for i, _ in enumerate(fh):
                self.assertEqual(self._file.lineno, i + 1)
        # test proper exiting
        self.assertIsNone(self._file._fh)
        try:
            with self._file.open("r") as fh:
                1 / 0
        except ZeroDivisionError:
            pass
        # test proper exiting with error
        self.assertIsNone(self._file._fh)

    def test_str(self):
        self.assertEqual(str(self._file), self._file_name)

    def test_read(self):
        with self._file.open("r") as fh:
            contents = fh.read()
        self.assertEqual(self._file.lineno, contents.count("\n") + 1)

    def test_readline(self):
        with self._file.open("r") as fh:
            i = 0
            line = "hi"
            while line:
                line = fh.readline()
                print(line.encode())
                if line:
                    i += 1
                self.assertEqual(self._file.lineno, i + 1)

    def test_write(self):
        out = "bar.imcnp"
        try:
            test = MCNP_InputFile(out)
            with test.open("w") as fh:
                fh.write("hi\nbar\n")
                self.assertEqual(test.lineno, 3)
            with test.open("r") as fh:
                contents = fh.read()
                self.assertIn("hi", contents)
        finally:
            if os.path.exists(out):
                os.remove(out)



# Copyright 2024, Battelle Energy Alliance, LLC All Rights Reserved.
BAD_INPUTS = {
    "test_excess_mt.imcnp",
    "test_vol_redundant.imcnp",
    "testVerticalMode.imcnp",
    "test_missing_mat_for_mt.imcnp",
    "test_imp_redundant.imcnp",
    "test_broken_cell_surf_link.imcnp",
    "test_broken_surf_link.imcnp",
    "test_broken_transform_link.imcnp",
    "test_broken_mat_link.imcnp",
    "test_broken_complement.imcnp",
    "test_bad_syntax.imcnp",
    "test_extra_params.imcnp",
    "test_extra_data_param.imcnp",
    "test_extra_data_imp.imcnp",
    "number_conflict_pin_cell.imcnp",
}
IGNORE_FILES = {
    "testReadRec1.imcnp",
    "testReadRec2.imcnp",
    "testReadRec3.imcnp",
    "testReadTarget.imcnp",
    "bad_encoding.imcnp",
    "unicode.imcnp",
}

BAD_ENCODING_FILES = {
    "bad_encoding.imcnp",
    "unicode.imcnp",
}



# Copyright 2024, Battelle Energy Alliance, LLC All Rights Reserved.
from montepy.data_inputs.element import Element
from montepy.errors import *
from montepy.input_parser.syntax_node import ValueNode


class Isotope:
    """
    A class to represent an MCNP isotope

    :param ZAID: the MCNP isotope identifier
    :type ZAID: str
    """

    #                   Cl-52      Br-101     Xe-150      Os-203    Cm-251     Og-296
    _BOUNDING_CURVE = [(17, 52), (35, 101), (54, 150), (76, 203), (96, 251), (118, 296)]
    """
    Points on bounding curve for determining if "valid" isotope
    """

    def __init__(self, ZAID="", node=None):
        if node is not None and isinstance(node, ValueNode):
            self._tree = node
            ZAID = node.value
        if "." in ZAID:
            parts = ZAID.split(".")
            try:
                assert len(parts) == 2
                int(parts[0])
            except (AssertionError, ValueError) as e:
                raise ValueError(f"ZAID: {ZAID} could not be parsed as a valid isotope")
            self._ZAID = parts[0]
            self.__parse_zaid()
            self._library = parts[1]
        else:
            raise ValueError(f"ZAID: {ZAID} could not be parsed as a valid isotope")

    def __parse_zaid(self):
        """
        Parses the ZAID fully including metastable isomers.

        See Table 3-32 of LA-UR-17-29881

        """

        def is_probably_an_isotope(Z, A):
            for lim_Z, lim_A in self._BOUNDING_CURVE:
                if Z <= lim_Z:
                    if A <= lim_A:
                        return True
                    else:
                        return False
                else:
                    continue
            # if you are above Lv it's probably legit.
            return True

        ZAID = int(self._ZAID)
        self._Z = int(ZAID / 1000)
        self._element = Element(self.Z)
        A = int(ZAID % 1000)
        if not is_probably_an_isotope(self.Z, A):
            self._is_metastable = True
            true_A = A - 300
            # only m1,2,3,4 allowed
            found = False
            for i in range(1, 5):
                true_A -= 100
                # assumes that can only vary 40% from A = 2Z
                if is_probably_an_isotope(self.Z, true_A):
                    found = True
                    break
            if found:
                self._meta_state = i
                self._A = true_A
            else:
                raise ValueError(
                    f"ZAID: {ZAID} cannot be parsed as a valid metastable isomer. "
                    "Only isomeric state 1 - 4 are allowed"
                )

        else:
            self._is_metastable = False
            self._meta_state = None
            self._A = A

    @property
    def ZAID(self):
        """
        The ZZZAAA identifier following MCNP convention

        :rtype: int
        """
        # if this is made mutable this cannot be user provided, but must be calculated.
        return self._ZAID

    @property
    def Z(self):
        """
        The Z number for this isotope.

        :returns: the atomic number.
        :rtype: int
        """
        return self._Z

    @property
    def A(self):
        """
        The A number for this isotope.

        :returns: the isotope's mass.
        :rtype: int
        """
        return self._A

    @property
    def element(self):
        """
        The base element for this isotope.

        :returns: The element for this isotope.
        :rtype: Element
        """
        return self._element

    @property
    def is_metastable(self):
        """
        Whether or not this is a metastable isomer.

        :returns: boolean of if this is metastable.
        :rtype: bool
        """
        return self._is_metastable

    @property
    def meta_state(self):
        """
        If this is a metastable isomer, which state is it?

        Can return values in the range [1,4] (or None). The exact state
        number is decided by who made the ACE file for this, and not quantum mechanics.
        Convention states that the isomers should be numbered from lowest to highest energy.

        :returns: the metastable isomeric state of this "isotope" in the range [1,4], or None
                if this is a ground state isomer.
        :rtype: int
        """
        return self._meta_state

    @property
    def library(self):
        """
         The MCNP library identifier e.g. 80c

        :rtype: str
        """
        return self._library

    @library.setter
    def library(self, library):
        if not isinstance(library, str):
            raise TypeError("library must be a string")
        self._library = library

    def __str__(self):
        return f"{self.element.symbol:>2}-{self.A:<3} ({self._library})"

    def mcnp_str(self):
        """
        Returns an MCNP formatted representation.

        E.g., 1001.80c

        :returns: a string that can be used in MCNP
        :rtype: str
        """
        return f"{self.ZAID}.{self.library}"

    def get_base_zaid(self):
        """
        Get the ZAID identifier of the base isotope this is an isomer of.

        This is mostly helpful for working with metastable isomers.

        :returns: the mcnp ZAID of the ground state of this isotope.
        :rtype: int
        """
        return self.Z * 1000 + self.A

    def __repr__(self):
        return f"ZAID={self.ZAID}, Z={self.Z}, A={self.A}, element={self.element}, library={self.library}"

    def __hash__(self):
        return hash(self._ZAID)

    def __lt__(self, other):
        return int(self.ZAID) < int(other.ZAID)

    def __format__(self, format_str):
        return str(self).__format__(format_str)



# Copyright 2024, Battelle Energy Alliance, LLC All Rights Reserved.
from abc import ABC, abstractmethod
import collections
import copy
import enum
import math

from montepy import input_parser
from montepy import constants
from montepy.constants import rel_tol, abs_tol
from montepy.errors import *
from montepy.input_parser.shortcuts import Shortcuts
from montepy.geometry_operators import Operator
from montepy.particle import Particle
from montepy.utilities import fortran_float
import re
import warnings


class SyntaxNodeBase(ABC):
    """
    A base class for all syntax nodes.

    A syntax node is any component of the syntax tree
    for a parsed input.

    .. versionadded:: 0.2.0
        This was added with the major parser rework.

    :param name: a name for labeling this node.
    :type name: str
    """

    def __init__(self, name):
        self._name = name
        self._nodes = []

    def append(self, node):
        """
        Append the node to this node.

        :param node: node
        :type node: SyntaxNodeBase, str, None
        """
        self._nodes.append(node)

    @property
    def nodes(self):
        """
        The children nodes of this node.

        :returns: a list of the nodes.
        :rtype: list
        """
        return self._nodes

    def __len__(self):
        return len(self.nodes)

    @property
    def name(self):
        """
        The name for the node.

        :returns: the node's name.
        :rtype: str
        """
        return self._name

    @name.setter
    def name(self, name):
        if not isinstance(name, str):
            raise TypeError("Name must be a string")
        self._name = name

    @abstractmethod
    def format(self):
        """
        Generate a string representing the tree's current state.

        :returns: the MCNP representation of the tree's current state.
        :rtype: str
        """
        pass

    @property
    @abstractmethod
    def comments(self):
        """
        A generator of all comments contained in this tree.

        :returns: the comments in the tree.
        :rtype: Generator
        """
        pass

    def get_trailing_comment(self):
        """
        Get the trailing comments if any.

        :returns: The trailing comments of this tree.
        :rtype: list
        """
        if len(self.nodes) == 0:
            return
        tail = self.nodes[-1]
        if isinstance(tail, SyntaxNodeBase):
            return tail.get_trailing_comment()

    def _delete_trailing_comment(self):
        """
        Deletes the trailing comment if any.
        """
        if len(self.nodes) == 0:
            return
        tail = self.nodes[-1]
        if isinstance(tail, SyntaxNodeBase):
            tail._delete_trailing_comment()


class SyntaxNode(SyntaxNodeBase):
    """
    A general syntax node for handling inner tree nodes.

    This is a generalized wrapper for a dictionary.
    The order of the dictionary is significant.

    This does behave like a dict for collecting items. e.g.,

    .. code-block:: python

        value = syntax_node["start_pad"]
        if key in syntax_node:
            pass

    .. versionadded:: 0.2.0
        This was added with the major parser rework.

    :param name: a name for labeling this node.
    :type name: str
    :param parse_dict: the dictionary of the syntax tree nodes.
    :type parse_dict: dict
    """

    def __init__(self, name, parse_dict):
        super().__init__(name)
        self._name = name
        self._nodes = parse_dict

    def __getitem__(self, key):
        return self.nodes[key]

    def __contains__(self, key):
        return key in self.nodes

    def get_value(self, key):
        """
        Get a value from the syntax tree.

        :param key: the key for the item to get.
        :type key: str
        :returns: the node in the syntax tree.
        :rtype: SyntaxNodeBase
        :raises KeyError: if key is not in SyntaxNode
        """
        temp = self.nodes[key]
        if isinstance(temp, ValueNode):
            return temp.value
        else:
            raise KeyError(f"{key} is not a value leaf node")

    def __str__(self):
        return f"(Node: {self.name}: {self.nodes})"

    def __repr__(self):
        return str(self)

    def format(self):
        ret = ""
        for node in self.nodes.values():
            if isinstance(node, ValueNode):
                if node.value is not None:
                    ret += node.format()
            else:
                ret += node.format()
        return ret

    @property
    def comments(self):
        for node in self.nodes.values():
            yield from node.comments

    def get_trailing_comment(self):
        if len(self.nodes) == 0:
            return
        tail = next(reversed(self.nodes.items()))
        return tail[1].get_trailing_comment()

    def _delete_trailing_comment(self):
        tail = next(reversed(self.nodes.items()))
        tail[1]._delete_trailing_comment()


class GeometryTree(SyntaxNodeBase):
    """
    A syntax tree that is a binary tree for representing CSG geometry logic.

    .. versionadded:: 0.2.0
        This was added with the major parser rework.

    :param name: a name for labeling this node.
    :type name: str
    :param tokens: The nodes that are in the tree.
    :type tokens: dict
    :param op: The string representation of the Operator to use.
    :type op: str
    :param left: the node of the left side of the binary tree.
    :type left: GeometryTree, ValueNode
    :param right: the node of the right side of the binary tree.
    :type right: GeometryTree, ValueNode
    """

    def __init__(self, name, tokens, op, left, right=None):
        super().__init__(name)
        assert all(list(map(lambda v: isinstance(v, SyntaxNodeBase), tokens.values())))
        self._nodes = tokens
        self._operator = Operator(op)
        self._left_side = left
        self._right_side = right

    def __str__(self):
        return f"Geometry: ( {self._left_side} {self._operator} {self._right_side} )"

    def __repr__(self):
        return str(self)

    def format(self):
        ret = ""
        for node in self.nodes.values():
            ret += node.format()
        return ret

    @property
    def comments(self):
        for node in self.nodes.values():
            yield from node.comments

    @property
    def left(self):
        """
        The left side of the binary tree.

        :returns: the left node of the syntax tree.
        :rtype: GeometryTree, ValueNode
        """
        return self._left_side

    @property
    def right(self):
        """
        The right side of the binary tree.

        :returns: the right node of the syntax tree.
        :rtype: GeometryTree, ValueNode
        """
        return self._right_side

    @property
    def operator(self):
        """
        The operator used for the binary tree.

        :returns: the operator used.
        :rtype: Operator
        """
        return self._operator

    def __iter__(self):
        """
        Iterates over the leafs
        """
        self._iter_l_r = False
        self._iter_complete = False
        self._sub_iter = None
        return self

    def __next__(self):
        if self._iter_complete:
            raise StopIteration
        if not self._iter_l_r:
            node = self.left
        if self._iter_l_r and self.right is not None:
            node = self.right
        if isinstance(node, ValueNode):
            if not self._iter_l_r:
                if self.right is not None:
                    self._iter_l_r = True
                else:
                    self._iter_complete = True
            else:
                self._iter_complete = True
            return node
        if self._sub_iter is None:
            self._sub_iter = iter(node)
        try:
            return next(self._sub_iter)
        except StopIteration:
            self._sub_iter = None
            if not self._iter_l_r:
                if self.right is not None:
                    self._iter_l_r = True
                else:
                    self._iter_complete = True
            else:
                raise StopIteration
            return next(self)


class PaddingNode(SyntaxNodeBase):
    """
    A syntax tree node to represent a collection of sequential padding elements.

    .. versionadded:: 0.2.0
        This was added with the major parser rework.

    :param token: The first padding token for this node.
    :type token: str
    :param is_comment: If the token provided is a comment.
    :type is_comment: bool
    """

    def __init__(self, token=None, is_comment=False):
        super().__init__("padding")
        if token is not None:
            self.append(token, is_comment)

    def __str__(self):
        return f"(Padding, {self._nodes})"

    def __repr__(self):
        return str(self)

    def __iadd__(self, other):
        if not isinstance(other, type(self)):
            raise TypeError(f"Can only combine with PaddingNodes. {other} given.")
        self._nodes += other.nodes
        return self

    @property
    def value(self):
        """
        A string representation of the contents of this node.

        All of the padding will be combined into a single string.

        :returns: a string sequence of the padding.
        :rtype: str
        """
        return "".join([val.format() for val in self.nodes])

    def is_space(self, i):
        """
        Determine if the value at i is a space or not.

        .. note::
            the newline, ``\\n``, by itself is not considered a space.

        :param i: the index of the element to check.
        :type i: int
        :returns: true iff the padding at that node is only spaces that are not ``\\n``.
        :raises IndexError: if the index i is not in ``self.nodes``.
        """
        val = self.nodes[i]
        if not isinstance(val, str):
            return False
        return len(val.strip()) == 0 and val != "\n"

    def append(self, val, is_comment=False):
        """
        Append the node to this node.

        :param node: node
        :type node: str, CommentNode
        :param is_comment: whether or not the node is a comment.
        :type is_comment: bool
        """
        if is_comment:
            if not isinstance(val, CommentNode):
                val = CommentNode(val)
            self.nodes.append(val)
            return
        parts = val.split("\n")
        if len(parts) > 1:
            for part in parts[:-1]:
                if part:
                    self._nodes += [part, "\n"]
                else:
                    self._nodes.append("\n")
            if parts[-1]:
                self._nodes.append(parts[-1])
        else:
            self._nodes.append(val)

    def format(self):
        ret = ""
        for node in self.nodes:
            if isinstance(node, str):
                ret += node
            else:
                ret += node.format()
        return ret

    @property
    def comments(self):
        for node in self.nodes:
            if isinstance(node, CommentNode):
                yield node

    def _get_first_comment(self):
        """
        Get the first index that is a comment.

        :returns: the index of the first comment, if there is no comment then None.
        :rtype: int, None
        """
        for i, item in enumerate(self.nodes):
            if isinstance(item, CommentNode):
                return i
        return None

    def get_trailing_comment(self):
        i = self._get_first_comment()
        if i is not None:
            return self.nodes[i:]
        return None

    def _delete_trailing_comment(self):
        i = self._get_first_comment()
        if i is not None:
            del self._nodes[i:]

    def _grab_beginning_comment(self, extra_padding):
        """
        Consumes the provided comment, and moves it to the beginning of this node.

        :param extra_padding: the padding comment to add to the beginning of this padding.
        :type extra_padding: list
        """
        if extra_padding[-1] != "\n":
            extra_padding.append("\n")
        self._nodes = extra_padding + self.nodes

    def __eq__(self, other):
        if not isinstance(other, (type(self), str)):
            raise "PaddingNode can only be compared to PaddingNode or str"
        if isinstance(other, type(self)):
            other = other.format()
        return self.format() == other


class CommentNode(SyntaxNodeBase):
    """
    Object to represent a comment in an MCNP problem.

    .. versionadded:: 0.2.0
        This was added with the major parser rework.

    :param input: the token from the lexer
    :type input: Token
    """

    _MATCHER = re.compile(
        rf"""(?P<delim>
                (\s{{0,{constants.BLANK_SPACE_CONTINUE-1}}}C\s?)
                |(\$\s?)
             )
            (?P<contents>.*)""",
        re.I | re.VERBOSE,
    )
    """
    A re matcher to confirm this is a C style comment.
    """

    def __init__(self, input):
        super().__init__("comment")
        is_dollar, node = self._convert_to_node(input)
        self._is_dollar = is_dollar
        self._nodes = [node]

    def _convert_to_node(self, token):
        """
        Converts the token to a Syntax Node to store.

        :param token: the token to convert.
        :type token: str
        :returns: the SyntaxNode of the Comment.
        :rtype: SyntaxNode
        """
        if match := self._MATCHER.match(token):
            start = match["delim"]
            comment_line = match["contents"]
            is_dollar = "$" in start
        else:
            start = token
            comment_line = ""
            is_dollar = "$" in start
        return (
            is_dollar,
            SyntaxNode(
                "comment",
                {
                    "delimiter": ValueNode(start, str),
                    "data": ValueNode(comment_line, str),
                },
            ),
        )

    def append(self, token):
        """
        Append the comment token to this node.

        :param token: the comment token
        :type token: str
        """
        is_dollar, node = self._convert_to_node(token)
        if is_dollar or self._is_dollar:
            raise TypeError(
                f"Cannot append multiple comments to a dollar comment. {token} given."
            )
        self._nodes.append(node)

    @property
    def is_dollar(self):
        """
        Whether or not this CommentNode is a dollar sign ($) comment.

        :returns: True iff this is a dollar sign comment.
        :rtype: bool
        """
        return self._is_dollar

    @property
    def contents(self):
        """
        The contents of the comments without delimiters (i.e., $/C).

        :returns: String of the contents
        :rtype: str
        """
        return "\n".join([node["data"].value for node in self.nodes])

    def format(self):
        ret = ""
        for node in self.nodes:
            ret += node.format()
        return ret

    @property
    def comments(self):
        yield from [self]

    def __str__(self):
        return self.format()

    def __repr__(self):
        ret = f"COMMENT:\n"
        for node in self.nodes:
            ret += node.format()
        return ret


class ValueNode(SyntaxNodeBase):
    """
    A syntax node to represent the leaf node.

    This stores the original input token, the current value,
    and the possible associated padding.

    .. versionadded:: 0.2.0
        This was added with the major parser rework.

    :param token: the original token for the ValueNode.
    :type token: str
    :param token_type: the type for the ValueNode.
    :type token_type: class
    :param padding: the padding for this node.
    :type padding: PaddingNode
    :param never_pad: If true an ending space will never be added to this.
    :type never_pad: bool
    """

    _FORMATTERS = {
        float: {
            "value_length": 0,
            "precision": 5,
            "zero_padding": 0,
            "sign": "-",
            "divider": "e",
            "exponent_length": 0,
            "exponent_zero_pad": 0,
            "as_int": False,
            "int_tolerance": 1e-6,
            "is_scientific": True,
        },
        int: {"value_length": 0, "zero_padding": 0, "sign": "-"},
        str: {"value_length": 0},
    }
    """
    The default formatters for each type.
    """

    _SCIENTIFIC_FINDER = re.compile(
        r"""
            [+\-]?                      # leading sign if any
            (?P<significand>\d+\.*\d*)  # the actual number
            ((?P<e>[eE])                 # non-optional e with +/-
            [+\-]?|
            [+\-])                  #non-optional +/- if fortran float is used
            (?P<exponent>\d+)                    #exponent
        """,
        re.VERBOSE,
    )
    """
    A regex for finding scientific notation.
    """

    def __init__(self, token, token_type, padding=None, never_pad=False):
        super().__init__("")
        self._token = token
        self._type = token_type
        self._formatter = self._FORMATTERS[token_type].copy()
        self._is_neg_id = False
        self._is_neg_val = False
        self._og_value = None
        self._never_pad = never_pad
        if token is None:
            self._value = None
        elif isinstance(token, input_parser.mcnp_input.Jump):
            self._value = None
        elif token_type == float:
            self._value = fortran_float(token)
        elif token_type == int:
            self._value = int(token)
        else:
            self._value = token
        self._og_value = self.value
        self._padding = padding
        self._nodes = [self]
        self._is_reversed = False

    def _convert_to_int(self):
        """
        Converts a float ValueNode to an int ValueNode.
        """
        if self._type not in {float, int}:
            raise ValueError(f"ValueNode must be a float to convert to int")
        self._type = int
        if self._token is not None and not isinstance(
            self._token, input_parser.mcnp_input.Jump
        ):
            try:
                self._value = int(self._token)
            except ValueError as e:
                parts = self._token.split(".")
                if len(parts) > 1 and int(parts[1]) == 0:
                    self._value = int(parts[0])
                else:
                    raise e
        self._formatter = self._FORMATTERS[int].copy()

    def _convert_to_enum(
        self, enum_class, allow_none=False, format_type=str, switch_to_upper=False
    ):
        """
        Converts the ValueNode to an Enum for allowed values.

        :param enum_class: the class for the enum to use.
        :type enum_class: Class
        :param allow_none: Whether or not to allow None as a value.
        :type allow_none: bool
        :param format_type: the base data type to format this ValueNode as.
        :type format_type: Class
        :param switch_to_upper: Whether or not to convert a string to upper case before convert to enum.
        :type switch_to_upper: bool
        """
        self._type = enum_class
        if switch_to_upper:
            value = self._value.upper()
        else:
            value = self._value
        if not (allow_none and self._value is None):
            self._value = enum_class(value)
        self._formatter = self._FORMATTERS[format_type].copy()

    @property
    def is_negatable_identifier(self):
        """
        Whether or not this value is a negatable identifier.

        Example use: the surface transform or periodic surface is switched based on positive
        or negative.

        This means:
            1. the ValueNode is an int.
            2. The ``value`` will always be positive.
            3. The ``is_negative`` property will be available.

        :returns: the state of this marker.
        :rtype: bool
        """
        return self._is_neg_id

    @is_negatable_identifier.setter
    def is_negatable_identifier(self, val):
        if val == True:
            self._convert_to_int()
            if self.value is not None:
                self._is_neg = self.value < 0
                self._value = abs(self._value)
            else:
                self._is_neg = None
        self._is_neg_id = val

    @property
    def is_negatable_float(self):
        """
        Whether or not this value is a negatable float.

        Example use: cell density.

        This means:
            1. the ValueNode is an int.
            2. The ``value`` will always be positive.
            3. The ``is_negative`` property will be available.

        :returns: the state of this marker.
        :rtype: bool
        """
        return self._is_neg_val

    @is_negatable_float.setter
    def is_negatable_float(self, val):
        if val == True:
            if self.value is not None:
                self._is_neg = self.value < 0
                self._value = abs(self._value)
            else:
                self._is_neg = None
        self._is_neg_val = val

    @property
    def is_negative(self):
        """
        Whether or not this value is negative.

        If neither :func:`is_negatable_float` or :func:`is_negatable_identifier` is true
        then this will return ``None``.

        :returns: true if this value is negative (either in input or through state).
        :rtype: bool, None
        """
        if self.is_negatable_identifier or self.is_negatable_float:
            return self._is_neg

    @is_negative.setter
    def is_negative(self, val):
        if self.is_negatable_identifier or self.is_negatable_float:
            self._is_neg = val

    def _reverse_engineer_formatting(self):
        """
        Tries its best to figure out and update the formatter based on the token's format.
        """
        if not self._is_reversed and self._token is not None:
            self._is_reversed = True
            token = self._token
            if isinstance(token, input_parser.mcnp_input.Jump):
                token = "J"
            if isinstance(token, (int, float)):
                token = str(token)
            self._formatter["value_length"] = len(token)
            if self.padding:
                if self.padding.is_space(0):
                    self._formatter["value_length"] += len(self.padding.nodes[0])

            if self._type == float or self._type == int:
                no_zero_pad = token.lstrip("0+-")
                length = len(token)
                delta = length - len(no_zero_pad)
                if token.startswith("+") or token.startswith("-"):
                    delta -= 1
                    if token.startswith("+"):
                        self._formatter["sign"] = "+"
                    if token.startswith("-"):
                        self._formatter["sign"] = " "
                if delta > 0:
                    self._formatter["zero_padding"] = length
                if self._type == float:
                    self._reverse_engineer_float()

    def _reverse_engineer_float(self):
        token = self._token
        if isinstance(token, float):
            token = str(token)
        if isinstance(token, input_parser.mcnp_input.Jump):
            token = "J"
        if match := self._SCIENTIFIC_FINDER.match(token):
            groups = match.groupdict(default="")
            self._formatter["is_scientific"] = True
            significand = groups["significand"]
            self._formatter["divider"] = groups["e"]
            # extra space for the "e" in scientific and... stuff
            self._formatter["zero_padding"] += 4
            exponent = groups["exponent"]
            temp_exp = exponent.lstrip("0")
            if exponent != temp_exp:
                self._formatter["exponent_length"] = len(exponent)
                self._formatter["exponent_zero_pad"] = len(exponent)
        else:
            self._formatter["is_scientific"] = False
            significand = token
        parts = significand.split(".")
        if len(parts) == 2:
            precision = len(parts[1])
        else:
            precision = self._FORMATTERS[float]["precision"]
            self._formatter["as_int"] = True

        self._formatter["precision"] = precision

    def _can_float_to_int_happen(self):
        """
        Checks if you can format a floating point as an int.

        E.g., 1.0 -> 1

        Considers if this was done in the input, and if the value is close to the int value.

        :rtype: bool.
        """
        if self._type != float or not self._formatter["as_int"]:
            return False
        nearest_int = round(self.value)
        if not math.isclose(nearest_int, self.value, rel_tol=rel_tol, abs_tol=abs_tol):
            return False
        return True

    @property
    def _print_value(self):
        """
        The print version of the value.

        This takes a float/int that is negatable, and negates it
        based on the ``is_negative`` value.

        :rtype: int, float
        """
        if self._type in {int, float} and self.is_negative:
            return -self.value
        return self.value

    @property
    def _value_changed(self):
        """
        Checks if the value has changed at all from first parsing.

        Used to shortcut formatting and reverse engineering.

        :rtype: bool
        """
        if self.value is None and self._og_value is None:
            return False
        if self.value is None or self._og_value is None:
            return True
        if self._type in {float, int}:
            return not math.isclose(
                self._print_value, self._og_value, rel_tol=rel_tol, abs_tol=abs_tol
            )
        return self.value != self._og_value

    def format(self):
        if not self._value_changed:
            return f"{self._token}{self.padding.format() if self.padding else ''}"
        if self.value is None:
            return ""
        self._reverse_engineer_formatting()
        if issubclass(self.type, enum.Enum):
            value = self.value.value
        else:
            value = self._print_value
        if self._type == int or self._can_float_to_int_happen():
            temp = "{value:0={sign}{zero_padding}d}".format(
                value=int(value), **self._formatter
            )
        elif self._type == float:
            # default to python general if new value
            if not self._is_reversed:
                temp = "{value:0={sign}{zero_padding}.{precision}g}".format(
                    value=value, **self._formatter
                )
            elif self._formatter["is_scientific"]:
                temp = "{value:0={sign}{zero_padding}.{precision}e}".format(
                    value=value, **self._formatter
                )
                temp = temp.replace("e", self._formatter["divider"])
                temp_match = self._SCIENTIFIC_FINDER.match(temp)
                exponent = temp_match.group("exponent")
                start, end = temp_match.span("exponent")
                new_exp_temp = "{value:0={zero_padding}d}".format(
                    value=int(exponent),
                    zero_padding=self._formatter["exponent_zero_pad"],
                )
                new_exp = "{temp:<{value_length}}".format(
                    temp=new_exp_temp, value_length=self._formatter["exponent_length"]
                )
                temp = temp[0:start] + new_exp + temp[end:]
            elif self._formatter["as_int"]:
                temp = "{value:0={sign}0{zero_padding}g}".format(
                    value=value, **self._formatter
                )
            else:
                temp = "{value:0={sign}0{zero_padding}.{precision}f}".format(
                    value=value, **self._formatter
                )
        else:
            temp = str(value)
        if self.padding:
            if self.padding.is_space(0):
                # if there was and end space, and we ran out of space, and there isn't
                # a saving space later on
                if len(temp) >= self._formatter["value_length"] and not (
                    len(self.padding) > 1
                    and (self.padding.is_space(1) or self.padding.nodes[1] == "\n")
                ):
                    pad_str = " "
                else:
                    pad_str = ""
                extra_pad_str = "".join([x.format() for x in self.padding.nodes[1:]])
            else:
                pad_str = ""
                extra_pad_str = "".join([x.format() for x in self.padding.nodes])
        else:
            pad_str = ""
            extra_pad_str = ""
        buffer = "{temp:<{value_length}}{padding}".format(
            temp=temp, padding=pad_str, **self._formatter
        )
        if len(buffer) > self._formatter["value_length"] and self._token is not None:
            warning = LineExpansionWarning(
                f"The value has expanded, and may change formatting. The original value was {self._token}, new value is {temp}."
            )
            warning.cause = "value"
            warning.og_value = self._token
            warning.new_value = temp
            warnings.warn(
                warning,
                stacklevel=2,
            )
        return buffer + extra_pad_str

    @property
    def comments(self):
        if self.padding is not None:
            yield from self.padding.comments
        else:
            yield from []

    def get_trailing_comment(self):
        if self.padding is None:
            return
        return self.padding.get_trailing_comment()

    def _delete_trailing_comment(self):
        if self.padding is None:
            return
        self.padding._delete_trailing_comment()

    @property
    def padding(self):
        """
        The padding if any for this ValueNode.

        :returns: the padding if any.
        :rtype: PaddingNode
        """
        return self._padding

    @padding.setter
    def padding(self, pad):
        self._padding = pad

    @property
    def type(self):
        """
        The data type for this ValueNode.

        Examples: float, int, str, Lattice

        :returns: the class for the value of this node.
        :rtype: Class
        """
        return self._type

    @property
    def token(self):
        """
        The original text (token) for this ValueNode.

        :returns: the original input.
        :rtype: str
        """
        return self._token

    def __str__(self):
        return f"(Value, {self._value}, padding: {self._padding})"

    def __repr__(self):
        return str(self)

    @property
    def value(self):
        """
        The current semantic value of this ValueNode.

        This is the parsed meaning in the type of ``self.type``,
        that can be updated. When this value is updated, next time format()
        is ran this value will be used.

        :returns: the node's value in type ``type``.
        :rtype: float, int, str, enum
        """
        return self._value

    @value.setter
    def value(self, value):
        if self.is_negative is not None and value is not None:
            value = abs(value)
        self._check_if_needs_end_padding(value)
        self._value = value

    def _check_if_needs_end_padding(self, value):
        if value is None or self.value is not None or self._never_pad:
            return
        # if not followed by a trailing space
        if self.padding is None:
            self.padding = PaddingNode(" ")

    def __eq__(self, other):
        if not isinstance(other, (type(self), str, int, float)):
            raise TypeError(
                f"ValueNode can't be equal to {type(other)} type. {other} given."
            )
        if isinstance(other, ValueNode):
            other_val = other.value
            if self.type != other.type:
                return False
        else:
            other_val = other
            if self.type != type(other):
                return False

        if self.type == float and self.value is not None and other_val is not None:
            return math.isclose(self.value, other_val, rel_tol=rel_tol, abs_tol=abs_tol)
        return self.value == other_val


class ParticleNode(SyntaxNodeBase):
    """
    A node to hold particles information in a :class:`ClassifierNode`.

    .. versionadded:: 0.2.0
        This was added with the major parser rework.

    :param name: the name for the node.
    :type name: str
    :param token: the original token from parsing
    :type token: str
    """

    _letter_finder = re.compile(r"([a-zA-Z])")

    def __init__(self, name, token):
        super().__init__(name)
        self._nodes = [self]
        self._token = token
        self._order = []
        classifier_chunks = token.replace(":", "").split(",")
        self._particles = set()
        self._formatter = {"upper": False}
        for chunk in classifier_chunks:
            part = Particle(chunk.upper())
            self._particles.add(part)
            self._order.append(part)

    @property
    def token(self):
        """
        The original text (token) for this ParticleNode.

        :returns: the original input.
        :rtype: str
        """
        return self._token

    @property
    def particles(self):
        """
        The particles included in this node.

        :returns: a set of the particles being used.
        :rtype: set
        """
        return self._particles

    @particles.setter
    def particles(self, values):
        if not isinstance(values, (list, set)):
            raise TypeError(f"Particles must be a set. {values} given.")
        for value in values:
            if not isinstance(value, Particle):
                raise TypeError(f"All particles must be a Particle. {value} given")
        if isinstance(values, list):
            self._order = values
            values = set(values)
        self._particles = values

    def add(self, value):
        """
        Add a particle to this node.

        :param value: the particle to add.
        :type value: Particle
        """
        if not isinstance(value, Particle):
            raise TypeError(f"All particles must be a Particle. {value} given")
        self._order.append(value)
        self._particles.add(value)

    def remove(self, value):
        """
        Remove a particle from this node.

        :param value: the particle to remove.
        :type value: Particle
        """
        if not isinstance(value, Particle):
            raise TypeError(f"All particles must be a Particle. {value} given")
        self._particles.remove(value)
        self._order.remove(value)

    @property
    def _particles_sorted(self):
        """
        The particles in this node ordered in a nice-ish way.

        Ordering:
            1. User input.
            2. Order of particles appended
            3. randomly at the end if all else fails.

        :rtype: list
        """
        ret = self._order
        ret_set = set(ret)
        remainder = self.particles - ret_set
        extras = ret_set - self.particles
        for straggler in sorted(remainder):
            ret.append(straggler)
        for useless in extras:
            ret.remove(useless)
        return ret

    def format(self):
        self._reverse_engineer_format()
        if self._formatter["upper"]:
            parts = [p.value.upper() for p in self._particles_sorted]
        else:
            parts = [p.value.lower() for p in self._particles_sorted]
        return f":{','.join(parts)}"

    def _reverse_engineer_format(self):
        total_match = 0
        upper_match = 0
        for match in self._letter_finder.finditer(self._token):
            if match:
                if match.group(0).isupper():
                    upper_match += 1
                total_match += 1
        if upper_match / total_match >= 0.5:
            self._formatter["upper"] = True

    @property
    def comments(self):
        yield from []

    def __repr__(self):
        return self.format()

    def __iter__(self):
        return iter(self.particles)


class ListNode(SyntaxNodeBase):
    """
    A node to represent a list of values.

    .. versionadded:: 0.2.0
        This was added with the major parser rework.

    :param name: the name of this node.
    :type name: str
    """

    def __init__(self, name):
        super().__init__(name)
        self._shortcuts = []

    def __repr__(self):
        return f"(list: {self.name}, {self.nodes})"

    def update_with_new_values(self, new_vals):
        """
        Update this list node with new values.

        This will first try to find if any shortcuts in the original input match up with
        the new values. If so it will then "zip" out those shortcuts to consume
        as many neighbor nodes as possible.
        Finally, the internal shortcuts, and list will be updated to reflect the new state.

        :param new_vals: the new values (a list of ValueNodes)
        :type new_vals: list
        """
        if not new_vals:
            self._nodes = []
            return
        new_vals_cache = {id(v): v for v in new_vals}
        # bind shortcuts to single site in new values
        for shortcut in self._shortcuts:
            for node in shortcut.nodes:
                if id(node) in new_vals_cache:
                    new_vals_cache[id(node)] = shortcut
                    shortcut.nodes.clear()
                    break
        self._expand_shortcuts(new_vals, new_vals_cache)
        self._shortcuts = []
        self._nodes = []
        for key, node in new_vals_cache.items():
            if isinstance(node, ShortcutNode):
                if (
                    len(self._shortcuts) > 0 and node is not self._shortcuts[-1]
                ) or len(self._shortcuts) == 0:
                    self._shortcuts.append(node)
                    self._nodes.append(node)
            else:
                self._nodes.append(node)
        end = self._nodes[-1]
        # pop off final shortcut if it's a jump the user left off
        if (
            isinstance(end, ShortcutNode)
            and end._type == Shortcuts.JUMP
            and len(end._original) == 0
        ):
            self._nodes.pop()
            self._shortcuts.pop()

    def _expand_shortcuts(self, new_vals, new_vals_cache):
        """
        Expands the existing shortcuts, and tries to "zip out" and consume their neighbors.

        :param new_vals: the new values.
        :type new_vals: list
        :param new_vals_cache: a dictionary mapping the id of the ValueNode to the ValueNode
            or ShortcutNode. This is ordered the same as ``new_vals``.
        :type new_vals_cache: dict
        """

        def try_expansion(shortcut, value):
            status = shortcut.consume_edge_node(
                value, 1, i == last_end + 1 and last_end != 0
            )
            if status:
                new_vals_cache[id(value)] = shortcut
            else:
                new_vals_cache[id(value)] = value
            return status

        def try_reverse_expansion(shortcut, i, last_end):
            if i > 1:
                for value in new_vals[i - 1 : last_end : -1]:
                    if shortcut.consume_edge_node(value, -1):
                        new_vals_cache[id(value)] = shortcut
                    else:
                        new_vals_cache[id(value)] = value
                        return

        def check_for_orphan_jump(value):
            """
            Checks if the current Jump is not tied to an existing Shortcut
            """
            nonlocal shortcut
            if value.value is None and shortcut is None:
                shortcut = ShortcutNode(p=None, short_type=Shortcuts.JUMP)
                if shortcut.consume_edge_node(value, 1):
                    new_vals_cache[id(value)] = shortcut

        shortcut = None
        last_end = 0
        for i, value in enumerate(new_vals_cache.values()):
            # found a new shortcut
            if isinstance(value, ShortcutNode):
                # shortcuts bumped up against each other
                if shortcut is not None:
                    last_end = i - 1
                shortcut = value
                if try_expansion(shortcut, new_vals[i]):
                    try_reverse_expansion(shortcut, i, last_end)
                else:
                    shortcut = None
            # otherwise it is actually a value to expand as well
            else:
                if shortcut is not None:
                    if not try_expansion(shortcut, new_vals[i]):
                        last_end = i - 1
                        shortcut = None
                        check_for_orphan_jump(new_vals[i])
                else:
                    check_for_orphan_jump(new_vals[i])

    def append(self, val):
        """
        Append the node to this node.

        :param node: node
        :type node: ValueNode, ShortcutNode
        """
        if isinstance(val, ShortcutNode):
            self._shortcuts.append(val)
        super().append(val)

    @property
    def comments(self):
        for node in self.nodes:
            yield from node.comments

    def format(self):
        ret = ""
        length = len(self.nodes)
        last_node = None
        for i, node in enumerate(self.nodes):
            # adds extra padding
            if (
                isinstance(node, ValueNode)
                and node.padding is None
                and i < length - 1
                and not isinstance(self.nodes[i + 1], PaddingNode)
            ):
                node.padding = PaddingNode(" ")
            if isinstance(last_node, ShortcutNode) and isinstance(node, ShortcutNode):
                ret += node.format(last_node)
            else:
                ret += node.format()
            last_node = node
        return ret

    def __iter__(self):
        for node in self.nodes:
            if isinstance(node, ShortcutNode):
                yield from node.nodes
            else:
                yield node

    def __getitem__(self, indx):
        if isinstance(indx, slice):
            return self.__get_slice(indx)
        if indx >= 0:
            for i, item in enumerate(self):
                if i == indx:
                    return item
        else:
            items = list(self)
            return items[indx]
        raise IndexError(f"{indx} not in ListNode")

    def __get_slice(self, i: slice):
        """
        Helper function for __getitem__ with slices.
        """
        rstep = i.step if i.step is not None else 1
        rstart = i.start
        rstop = i.stop
        if rstep < 0:  # Backwards
            if rstart is None:
                rstart = len(self.nodes) - 1
            if rstop is None:
                rstop = 0
            rstop -= 1
        else:  # Forwards
            if rstart is None:
                rstart = 0
            if rstop is None:
                rstop = len(self.nodes) - 1
            rstop += 1
        buffer = []
        allowed_indices = range(rstart, rstop, rstep)
        for i, item in enumerate(self):
            if i in allowed_indices:
                buffer.append(item)
        ret = ListNode(f"{self.name}_slice")
        if rstep < 0:
            buffer.reverse()
        for val in buffer:
            ret.append(val)
        return ret

    def remove(self, obj):
        """
        Removes the given object from this list.

        :param obj: the object to remove.
        :type obj: ValueNode
        """
        self.nodes.remove(obj)

    def __eq__(self, other):
        if not isinstance(other, (type(self), list)):
            raise TypeError(
                f"ListNode can only be compared to a ListNode or List. {other} given."
            )
        if len(self) != len(other):
            return False
        for lhs, rhs in zip(self, other):
            if lhs != rhs:
                return False
        return True


class IsotopesNode(SyntaxNodeBase):
    """
    A node for representing isotopes and their concentration.

    This stores a list of tuples of ZAIDs and concentrations.

    .. versionadded:: 0.2.0
        This was added with the major parser rework.

    :param name: a name for labeling this node.
    :type name: str
    """

    def __init__(self, name):
        super().__init__(name)

    def append(self, isotope_fraction):
        """
        Append the node to this node.

        :param isotope_fraction: the isotope_fraction to add. This must be a tuple from
            A Yacc production. This will consist of: the string identifying the Yacc production,
            a ValueNode that is the ZAID, and a ValueNode of the concentration.
        :type isotope_fraction: tuple
        """
        isotope, concentration = isotope_fraction[1:3]
        self._nodes.append((isotope, concentration))

    def format(self):
        ret = ""
        for isotope, concentration in self.nodes:
            ret += isotope.format() + concentration.format()
        return ret

    def __repr__(self):
        return f"(Isotopes: {self.nodes})"

    def __iter__(self):
        return iter(self.nodes)

    @property
    def comments(self):
        for node in self.nodes:
            for value in node:
                yield from value.comments

    def get_trailing_comment(self):
        tail = self.nodes[-1]
        tail = tail[1]
        return tail.get_trailing_comment()

    def _delete_trailing_comment(self):
        tail = self.nodes[-1]
        tail = tail[1]
        tail._delete_trailing_comment()


class ShortcutNode(ListNode):
    """
    A node that pretends to be a :class:`ListNode` but is actually representing a shortcut.

    This takes the shortcut tokens, and expands it into their "virtual" values.

    .. versionadded:: 0.2.0
        This was added with the major parser rework.

    :param p: the parsing object to parse.
    :type p: sly.yacc.YaccProduction
    :param short_type: the type of the shortcut.
    :type short_type: Shortcuts
    """

    _shortcut_names = {
        ("REPEAT", "NUM_REPEAT"): Shortcuts.REPEAT,
        ("JUMP", "NUM_JUMP"): Shortcuts.JUMP,
        ("INTERPOLATE", "NUM_INTERPOLATE"): Shortcuts.INTERPOLATE,
        ("LOG_INTERPOLATE", "NUM_LOG_INTERPOLATE"): Shortcuts.LOG_INTERPOLATE,
        ("MULTIPLY", "NUM_MULTIPLY"): Shortcuts.MULTIPLY,
    }
    _num_finder = re.compile(r"\d+")

    def __init__(self, p=None, short_type=None):
        self._type = None
        self._end_pad = None
        self._nodes = collections.deque()
        self._original = []
        self._full = False
        self._num_node = ValueNode(None, float, never_pad=True)
        if p is not None:
            for search_strs, shortcut in self._shortcut_names.items():
                for search_str in search_strs:
                    if hasattr(p, search_str):
                        super().__init__(search_str.lower())
                        self._type = shortcut
            if self._type is None:
                raise ValueError("must use a valid shortcut")
            self._original = list(p)
            if self._type == Shortcuts.REPEAT:
                self._expand_repeat(p)
            elif self._type == Shortcuts.MULTIPLY:
                self._expand_multiply(p)
            elif self._type == Shortcuts.JUMP:
                self._expand_jump(p)
            elif self._type in {Shortcuts.INTERPOLATE, Shortcuts.LOG_INTERPOLATE}:
                self._expand_interpolate(p)
        elif short_type is not None:
            if not isinstance(short_type, Shortcuts):
                raise TypeError(f"Shortcut type must be Shortcuts. {short_type} given.")
            self._type = short_type
            if self._type in {
                Shortcuts.INTERPOLATE,
                Shortcuts.LOG_INTERPOLATE,
                Shortcuts.JUMP,
                Shortcuts.JUMP,
            }:
                self._num_node = ValueNode(None, int, never_pad=True)
            self._end_pad = PaddingNode(" ")

    @property
    def end_padding(self):
        """
        The padding at the end of this shortcut.

        :rtype: PaddingNode
        """
        return self._end_pad

    @end_padding.setter
    def end_padding(self, padding):
        if not isinstance(padding, PaddingNode):
            raise TypeError(
                f"End padding must be of type PaddingNode. {padding} given."
            )
        self._end_pad = padding

    def __repr__(self):
        return f"(shortcut:{self._type}: {self.nodes})"

    def _get_last_node(self, p):
        last = p[0]
        if isinstance(last, ValueNode):
            return collections.deque([last])
        return collections.deque()

    def _expand_repeat(self, p):
        self._nodes = self._get_last_node(p)
        repeat = p[1]
        try:
            repeat_num_str = repeat.lower().replace("r", "")
            repeat_num = int(repeat_num_str)
            self._num_node = ValueNode(repeat_num, int, never_pad=True)
        except ValueError:
            repeat_num = 1
            self._num_node = ValueNode(None, int, never_pad=True)
        if isinstance(p[0], ValueNode):
            last_val = p[0]
        else:
            if isinstance(p[0], GeometryTree):
                last_val = list(p[0])[-1]
            else:
                last_val = p[0].nodes[-1]
        if last_val.value is None:
            raise ValueError(f"Repeat cannot follow a jump. Given: {list(p)}")
        self._nodes += [copy.deepcopy(last_val) for i in range(repeat_num)]

    def _expand_multiply(self, p):
        self._nodes = self._get_last_node(p)
        mult_str = p[1].lower().replace("m", "")
        mult_val = fortran_float(mult_str)
        self._num_node = ValueNode(mult_str, float, never_pad=True)
        if isinstance(p[0], ValueNode):
            last_val = self.nodes[-1]
        else:
            last_val = p[0].nodes[-1]
        if last_val.value is None:
            raise ValueError(f"Multiply cannot follow a jump. Given: {list(p)}")
        self._nodes.append(copy.deepcopy(last_val))
        self.nodes[-1].value *= mult_val

    def _expand_jump(self, p):
        try:
            jump_str = p[0].lower().replace("j", "")
            jump_num = int(jump_str)
            self._num_node = ValueNode(jump_str, int, never_pad=True)
        except ValueError:
            jump_num = 1
            self._num_node = ValueNode(None, int, never_pad=True)
        for i in range(jump_num):
            self._nodes.append(
                ValueNode(input_parser.mcnp_input.Jump(), float, never_pad=True)
            )

    def _expand_interpolate(self, p):
        if self._type == Shortcuts.LOG_INTERPOLATE:
            is_log = True
        else:
            is_log = False
        if hasattr(p, "geometry_term"):
            term = p.geometry_term
            if isinstance(term, GeometryTree):
                begin = list(term)[-1].value
            else:
                begin = term.value
            end = p.number_phrase.value
        else:
            if isinstance(p[0], ListNode):
                begin = p[0].nodes[-1].value
            else:
                begin = p[0].value
            end = p.number_phrase.value
        self._nodes = self._get_last_node(p)
        if begin is None:
            raise ValueError(f"Interpolates cannot follow a jump. Given: {list(p)}")
        match = self._num_finder.search(p[1])
        if match:
            number = int(match.group(0))
            self._num_node = ValueNode(match.group(0), int, never_pad=True)
        else:
            number = 1
            self._num_node = ValueNode(None, int, never_pad=True)
        if is_log:
            begin = math.log(begin, 10)
            end = math.log(end, 10)
        spacing = (end - begin) / (number + 1)
        for i in range(number):
            if is_log:
                new_val = 10 ** (begin + spacing * (i + 1))
            else:
                new_val = begin + spacing * (i + 1)
            self.append(ValueNode(str(new_val), float, never_pad=True))
        self._begin = begin
        self._end = end
        self._spacing = spacing
        self.append(p.number_phrase)

    def _can_consume_node(self, node, direction, last_edge_shortcut=False):
        """
        If it's possible to consume this node.

        :param node: the node to consume
        :type node: ValueNode
        :param direction: the direct to go in. Must be in {-1, 1}
        :type direction: int
        :param last_edge_shortcut: Whether or the previous node in the list was
            part of a different shortcut
        :type last_edge_shortcut: bool
        :returns: true it can be consumed.
        :rtype: bool
        """
        if self._type == Shortcuts.JUMP:
            if node.value is None:
                return True

        # REPEAT
        elif self._type == Shortcuts.REPEAT:
            if len(self.nodes) == 0 and node.value is not None:
                return True
            if direction == 1:
                edge = self.nodes[-1]
            else:
                edge = self.nodes[0]
            if edge.type != node.type or edge.value is None or node.value is None:
                return False
            if edge.type in {int, float} and math.isclose(
                edge.value, node.value, rel_tol=rel_tol, abs_tol=abs_tol
            ):
                return True
            elif edge.value == node.value:
                return True

        # INTERPOLATE
        elif self._type in {Shortcuts.INTERPOLATE, Shortcuts.LOG_INTERPOLATE}:
            return self._is_valid_interpolate_edge(node, direction)
        # Multiply can only ever have 1 value
        elif self._type == Shortcuts.MULTIPLY:
            # can't do a multiply with a Jump
            if node.value is None:
                return False
            if len(self.nodes) == 0:
                # clear out old state if needed
                self._full = False
                if last_edge_shortcut:
                    self._full = True
                return True
            if len(self.nodes) == 1 and not self._full:
                return True
        return False

    def _is_valid_interpolate_edge(self, node, direction):
        """
        Is a valid interpolation edge.

        :param node: the node to consume
        :type node: ValueNode
        :param direction: the direct to go in. Must be in {-1, 1}
        :type direction: int
        :returns: true it can be consumed.
        :rtype: bool
        """
        # kill jumps immediately
        if node.value is None:
            return False
        if len(self.nodes) == 0:
            new_val = self._begin if direction == 1 else self._end
            if self._type == Shortcuts.LOG_INTERPOLATE:
                new_val = 10**new_val
        else:
            edge = self.nodes[-1] if direction == 1 else self.nodes[0]
            edge = edge.value
            if self._type == Shortcuts.LOG_INTERPOLATE:
                edge = math.log(edge, 10)
                new_val = 10 ** (edge + direction * self._spacing)
            else:
                new_val = edge + direction * self._spacing
        return math.isclose(new_val, node.value, rel_tol=rel_tol, abs_tol=abs_tol)

    def consume_edge_node(self, node, direction, last_edge_shortcut=False):
        """
        Tries to consume the given edge.

        If it can be consumed the node is appended to the internal nodes.

        :param node: the node to consume
        :type node: ValueNode
        :param direction: the direct to go in. Must be in {-1, 1}
        :type direction: int
        :param last_edge_shortcut: Whether or the previous node in the list was
            part of a different shortcut
        :type last_edge_shortcut: bool
        :returns: True if the node was consumed.
        :rtype: bool
        """
        if self._can_consume_node(node, direction, last_edge_shortcut):
            if direction == 1:
                self._nodes.append(node)
            else:
                self._nodes.appendleft(node)
            return True
        return False

    def format(self, leading_node=None):
        if self._type == Shortcuts.JUMP:
            temp = self._format_jump()
        # repeat
        elif self._type == Shortcuts.REPEAT:
            temp = self._format_repeat(leading_node)
        elif self._type == Shortcuts.MULTIPLY:
            temp = self._format_multiply(leading_node)
        elif self._type in {Shortcuts.INTERPOLATE, Shortcuts.LOG_INTERPOLATE}:
            temp = self._format_interpolate(leading_node)
        if self.end_padding:
            pad_str = self.end_padding.format()
        else:
            pad_str = ""
        return f"{temp}{pad_str}"

    def _format_jump(self):
        num_jumps = len(self.nodes)
        if num_jumps == 0:
            return ""
        if len(self._original) > 0 and "j" in self._original[0]:
            j = "j"
        else:
            j = "J"
        length = len(self._original)
        self._num_node.value = num_jumps
        if num_jumps == 1 and (
            length == 0 or (length > 0 and "1" not in self._original[0])
        ):
            num_jumps = ""
        else:
            num_jumps = self._num_node

        return f"{num_jumps.format()}{j}"

    def _format_repeat(self, leading_node=None):
        if leading_node is not None:
            first_val = ""
            num_extra = 0
        else:
            first_val = self.nodes[0].format()
            num_extra = 1
        num_repeats = len(self.nodes) - num_extra
        self._num_node.value = num_repeats
        if len(self._original) >= 2 and "r" in self._original[1]:
            r = "r"
        else:
            r = "R"
        if (
            num_repeats == 1
            and len(self._original) >= 2
            and "1" not in self._original[1]
        ):
            num_repeats = ""
        else:
            num_repeats = self._num_node
        return f"{first_val}{num_repeats.format()}{r}"

    def _format_multiply(self, leading_node=None):
        if leading_node is not None:
            first_val = leading_node.nodes[-1]
            first_val_str = ""
        else:
            first_val = self.nodes[0]
            first_val_str = first_val
        if "M" in self._original[-1]:
            m = "M"
        else:
            m = "m"
        self._num_node.value = self.nodes[-1].value / first_val.value
        return f"{first_val_str.format()}{self._num_node.format()}{m}"

    def _format_interpolate(self, leading_node=None):
        if leading_node is not None:
            start = ""
            num_extra_nodes = 1
        else:
            start = self.nodes[0]
            num_extra_nodes = 2
        end = self.nodes[-1]
        num_interp = len(self.nodes) - num_extra_nodes
        self._num_node.value = num_interp
        interp = "I"
        can_match = False
        if len(self._original) > 0:
            if match := re.match(r"\d*(\w+)", self._original[1]):
                can_match = True
                interp = match.group(1)
        if not can_match:
            if self._type == Shortcuts.LOG_INTERPOLATE:
                interp = "ILOG"
        if (
            num_interp == 1
            and len(self._original) >= 2
            and "1" not in self._original[1]
        ):
            num_interp = ""
        else:
            num_interp = self._num_node
        if len(self._original) >= 3:
            padding = self._original[2]
        else:
            padding = PaddingNode(" ")
        return f"{start.format()}{num_interp.format()}{interp}{padding.format()}{end.format()}"


class ClassifierNode(SyntaxNodeBase):
    """
    A node to represent the classifier for a :class:`montepy.data_input.DataInput`

    .. versionadded:: 0.2.0
        This was added with the major parser rework.

    e.g., represents ``M4``, ``F104:n,p``, ``IMP:n,e``.
    """

    def __init__(self):
        super().__init__("classifier")
        self._prefix = None
        self._number = None
        self._particles = None
        self._modifier = None
        self._padding = None
        self._nodes = []

    @property
    def prefix(self):
        """
        The prefix for the classifier.

        That is the string that tells what type of input this is.

        E.g.: ``M`` in ``M4`` or ``IMP`` in ``IMP:n``.

        :returns: the prefix
        :rtype: ValueNode
        """
        return self._prefix

    @prefix.setter
    def prefix(self, pref):
        self.append(pref)
        self._prefix = pref

    @property
    def number(self):
        """
        The number if any for the classifier.

        :returns: the number holder for this classifier.
        :rtype: ValueNode
        """
        return self._number

    @number.setter
    def number(self, number):
        self.append(number)
        self._number = number

    @property
    def particles(self):
        """
        The particles if any tied to this classifier.

        :returns: the particles used.
        :rtype: ParticleNode
        """
        return self._particles

    @particles.setter
    def particles(self, part):
        self.append(part)
        self._particles = part

    @property
    def modifier(self):
        """
        The modifier for this classifier if any.

        A modifier is a prefix character that changes the inputs behavior,
        e.g.: ``*`` or ``+``.

        :returns: the modifier
        :rtype: ValueNode
        """
        return self._modifier

    @modifier.setter
    def modifier(self, mod):
        self.append(mod)
        self._modifier = mod

    @property
    def padding(self):
        """
        The padding for this classifier.

        .. Note::
            None of the ValueNodes in this object should have padding.

        :returns: the padding after the classifier.
        :rtype: PaddingNode
        """
        return self._padding

    @padding.setter
    def padding(self, val):
        self.append(val)
        self._padding = val

    def format(self):
        if self.modifier:
            ret = self.modifier.format()
        else:
            ret = ""
        ret += self.prefix.format()
        if self.number:
            ret += self.number.format()
        if self.particles:
            ret += self.particles.format()
        if self.padding:
            ret += self.padding.format()
        return ret

    def __str__(self):
        return self.format()

    def __repr__(self):
        return (
            f"(Classifier: mod: {self.modifier}, prefix: {self.prefix}, "
            f"number: {self.number}, particles: {self.particles})"
        )

    @property
    def comments(self):
        if self.padding is not None:
            yield from self.padding.comments
        else:
            yield from []


class ParametersNode(SyntaxNodeBase):
    """
    A node to hold the parameters, key-value pairs, for this input.

    This behaves like a dictionary and is accessible by their key*

    .. versionadded:: 0.2.0
        This was added with the major parser rework.

    .. Note::
        How to access values.

        The internal dictionary doesn't use the full classifier directly,
        because some parameters should not be both allowed: e.g., ``fill`` and ``*fill``.
        The key is a string that is all lower case, and only uses the classifiers prefix,
        and particles.

        So to access a cell's fill information you would run:

        .. code-block:: python

            parameters["fill"]

        And to access the n,p importance:

        .. code-block:: python

            parameters["imp:n,p"]
    """

    def __init__(self):
        super().__init__("parameters")
        self._nodes = {}

    def append(self, val):
        """
        Append the node to this node.

        This takes a syntax node, which requires the keys:
            ``["classifier", "seperator", "data"]``

        :param val: the parameter to append.
        :type val: SyntaxNode
        """
        classifier = val["classifier"]
        key = (
            classifier.prefix.value
            + (str(classifier.particles) if classifier.particles else "")
        ).lower()
        if key in self._nodes:
            raise RedundantParameterSpecification(key, val)
        self._nodes[key] = val

    def __str__(self):
        return f"(Parameters, {self.nodes})"

    def __repr__(self):
        return str(self)

    def __getitem__(self, key):
        return self.nodes[key.lower()]

    def __contains__(self, key):
        return key.lower() in self.nodes

    def format(self):
        ret = ""
        for node in self.nodes.values():
            ret += node.format()
        return ret

    def get_trailing_comment(self):
        tail = next(reversed(self.nodes.items()))
        return tail[1].get_trailing_comment()

    def _delete_trailing_comment(self):
        tail = next(reversed(self.nodes.items()))
        tail[1]._delete_trailing_comment()

    @property
    def comments(self):
        for node in self.nodes.values():
            if isinstance(node, SyntaxNodeBase):
                yield from node.comments



import itertools
from unittest import TestCase
from tests import constants
import os
import subprocess


class TestChangeAsciiScript(TestCase):
    @classmethod
    def setUpClass(cls):
        new_files = {}
        for input_file in constants.BAD_ENCODING_FILES:
            new_files[input_file] = {}
            for flag in {"-w", "-d"}:
                new_file = f"{input_file}{flag}.imcnp"
                cls.run_script(
                    [flag, os.path.join("tests", "inputs", input_file), new_file]
                )
                new_files[input_file][flag] = new_file
        cls.files = new_files

    @classmethod
    def tearDownClass(cls):
        for group in cls.files.values():
            for file_name in group.values():
                try:
                    os.remove(file_name)
                except FileNotFoundError:
                    pass

    @staticmethod
    def run_script(args):
        return subprocess.run(
            ["python", os.path.join("montepy", "_scripts", "change_to_ascii.py")] + args
        )

    def test_delete_bad(self):
        for in_file in self.files:
            with open(os.path.join("tests", "inputs", in_file), "rb") as in_fh, open(
                self.files[in_file]["-d"], "rb"
            ) as out_fh:
                for in_line, out_line in zip(in_fh, out_fh):
                    try:
                        in_line.decode("ascii")
                        self.assertEqual(in_line, out_line)
                    except UnicodeError:
                        new_line = []
                        for char in in_line:
                            if char <= 128:
                                new_line.append(chr(char))
                        self.assertEqual("".join(new_line), out_line.decode("ascii"))

    def test_whitespace_bad(self):
        for in_file in self.files:
            with open(os.path.join("tests", "inputs", in_file), "rb") as in_fh, open(
                self.files[in_file]["-w"], "rb"
            ) as out_fh:
                for in_line, out_line in zip(in_fh, out_fh):
                    try:
                        in_line.decode("ascii")
                        self.assertEqual(in_line, out_line)
                    except UnicodeError:
                        new_line = []
                        try:
                            # try to change to utf-8
                            for char in in_line.decode():
                                if ord(char) <= 128:
                                    new_line.append(char)
                                else:
                                    new_line.append(" ")
                        except UnicodeError:
                            # try to go bit by bit
                            for char in in_line:
                                if char <= 128:
                                    new_line.append(chr(char))
                                else:
                                    new_line.append(" ")
                        self.assertEqual("".join(new_line), out_line.decode("ascii"))

    def test_bad_arguments(self):
        ret_code = self.run_script(
            [
                "-w",
                "-d",
                os.path.join("tests", "inputs", "bad_encode.imcnp", "foo.imcnp"),
            ]
        )
        self.assertNotEqual(ret_code.returncode, 0)



# Copyright 2024, Battelle Energy Alliance, LLC All Rights Reserved.
import copy
from unittest import TestCase

import numpy as np
import montepy
from montepy.data_inputs.transform import Transform
from montepy.errors import MalformedInputError
from montepy.input_parser.block_type import BlockType
from montepy.input_parser.mcnp_input import Input


class testTransformClass(TestCase):
    def test_transform_init(self):
        # test wrong ID finder
        with self.assertRaises(MalformedInputError):
            card = Input(["M20"], BlockType.DATA)
            Transform(card)
        # test that the minimum word requirement is set
        with self.assertRaises(MalformedInputError):
            card = Input(["TR5"], BlockType.DATA)
            Transform(card)
        # test that the transform is a valid number
        with self.assertRaises(MalformedInputError):
            card = Input(["TR1foo 0.0 0.0 0.0"], BlockType.DATA)
            Transform(card)

        # test invalid displacement vector
        with self.assertRaises(MalformedInputError):
            card = Input(["TR1 0.0 0.0 hi"], BlockType.DATA)
            Transform(card)
        # test invalid matrix
        with self.assertRaises(MalformedInputError):
            card = Input(["TR1 0.0 0.0 0.0 0.0 0.0 0.0 0.0 hi"], BlockType.DATA)
            Transform(card)
        # test invalid main to aux
        with self.assertRaises(MalformedInputError):
            card = Input(
                ["TR1 0.0 0.0 0.0 0.0 0.0 0.0 0.0 0.0 0.0 0.0 0.0 0.0 2"],
                BlockType.DATA,
            )
            Transform(card)
        # test that the transform has a number
        with self.assertRaises(MalformedInputError):
            card = Input(["*TR 0.0 0.0 0.0"], BlockType.DATA)
            Transform(card)
        # test that the transform doesn't have a particle
        with self.assertRaises(MalformedInputError):
            card = Input(["TR5:n,p 0.0 0.0 0.0"], BlockType.DATA)
            Transform(card)

        # test vanilla case
        in_str = "tr5 " + "1.0 " * 3 + "0.0 " * 9
        card = Input([in_str], BlockType.DATA)
        transform = Transform(card)
        self.assertEqual(transform.number, 5)
        self.assertEqual(transform.old_number, 5)
        self.assertFalse(transform.is_in_degrees)
        self.assertTrue(transform.is_main_to_aux)
        self.assertEqual(len(transform.displacement_vector), 3)
        self.assertEqual(len(transform.rotation_matrix), 9)
        for component in transform.displacement_vector:
            self.assertEqual(component, 1.0)
        for component in transform.rotation_matrix:
            self.assertEqual(component, 0.0)
        # Test in degrees form
        in_str = "*tr5 " + "1.0 " * 3 + "0.0 " * 9
        card = Input([in_str], BlockType.DATA)
        transform = Transform(card)
        self.assertTrue(transform.is_in_degrees)
        # test bad displace
        in_str = "*tr5 " + "de " * 3
        card = Input([in_str], BlockType.DATA)
        with self.assertRaises(MalformedInputError):
            Transform(card)
        # test bad rotation
        in_str = "*tr5 " + "1.0 " * 3 + "foo "
        card = Input([in_str], BlockType.DATA)
        with self.assertRaises(MalformedInputError):
            Transform(card)

        # test main to aux
        in_str = "*tr5 " + "1.0 " * 3 + "0.0 " * 9 + " -1"
        card = Input([in_str], BlockType.DATA)
        transform = Transform(card)
        self.assertFalse(transform.is_main_to_aux)
        in_str = "*tr5 " + "1.0 " * 3 + "0.0 " * 9 + " 1"
        card = Input([in_str], BlockType.DATA)
        transform = Transform(card)
        self.assertTrue(transform.is_main_to_aux)
        in_str = "*tr5 " + "1.0 " * 3 + "0.0 " * 9 + " foo"
        card = Input([in_str], BlockType.DATA)
        with self.assertRaises(MalformedInputError):
            Transform(card)

        # test blank init
        transform = Transform()
        self.assertEqual(transform.number, -1)
        self.assertEqual(len(transform.displacement_vector), 0)
        self.assertEqual(len(transform.rotation_matrix), 0)
        self.assertTrue(not transform.is_in_degrees)
        self.assertTrue(transform.is_main_to_aux)

    def test_validate(self):
        transform = Transform()
        with self.assertRaises(montepy.errors.IllegalState):
            transform.validate()
        with self.assertRaises(montepy.errors.IllegalState):
            transform.format_for_mcnp_input((6, 2, 0))

    def test_transform_degrees_setter(self):
        in_str = "*tr5 " + "1.0 " * 3 + "0.0 " * 9 + " -1"
        card = Input([in_str], BlockType.DATA)
        transform = Transform(card)
        transform.is_in_degrees = False
        self.assertFalse(transform.is_in_degrees)
        with self.assertRaises(TypeError):
            transform.is_in_degrees = 1

    def test_transform_number_setter(self):
        in_str = "*tr5 " + "1.0 " * 3 + "0.0 " * 9 + " -1"
        card = Input([in_str], BlockType.DATA)
        transform = Transform(card)
        transform.number = 20
        self.assertEqual(transform.number, 20)
        with self.assertRaises(TypeError):
            transform.number = "hi"
        with self.assertRaises(ValueError):
            transform.number = -5

    def test_transform_displace_setter(self):
        in_str = "*tr5 " + "1.0 " * 3 + "0.0 " * 9 + " -1"
        card = Input([in_str], BlockType.DATA)
        transform = Transform(card)
        transform.displacement_vector = np.array([1.0, 1.0, 1.0])
        for component in transform.displacement_vector:
            self.assertEqual(component, 1.0)
        with self.assertRaises(TypeError):
            transform.displacement_vector = [1, 2]
        with self.assertRaises(ValueError):
            transform.displacement_vector = np.array([1.0])
        with self.assertRaises(ValueError):
            transform.displacement_vector = np.array([1.0] * 10)

    def test_tranform_rotation_setter(self):
        in_str = "*tr5 " + "1.0 " * 3 + "0.0 " * 9 + " -1"
        card = Input([in_str], BlockType.DATA)
        transform = Transform(card)
        transform.rotation_matrix = np.array([1.0] * 5)
        for component in transform.rotation_matrix:
            self.assertEqual(component, 1.0)
        with self.assertRaises(TypeError):
            transform.rotation_matrix = [1, 2]
        with self.assertRaises(ValueError):
            transform.rotation_matrix = np.array([1.0])
        with self.assertRaises(ValueError):
            transform.displacement_vector = np.array([1.0] * 10)

    def test_is_main_aux_setter(self):
        in_str = "*tr5 " + "1.0 " * 3 + "0.0 " * 9
        card = Input([in_str], BlockType.DATA)
        transform = Transform(card)
        transform.is_main_to_aux = False
        self.assertFalse(transform.is_main_to_aux)
        with self.assertRaises(TypeError):
            transform.is_main_to_aux = "hi"

    def test_transform_str(self):
        in_str = "*tr5 " + "0.0 " * 3 + "0.0 " * 9 + " -1"
        card = Input([in_str], BlockType.DATA)
        transform = Transform(card)
        answer = """TRANSFORM: 5
DISPLACE: [0. 0. 0.]
ROTATE: [0. 0. 0. 0. 0. 0. 0. 0. 0.]
MAIN_TO_AUX: False
"""
        self.assertEqual(answer, repr(transform))
        self.assertEqual("TRANSFORM: 5", str(transform))

    def test_transform_print_mcnp(self):
        in_str = "tr5 " + "0.0 " * 3 + "0.0 " * 9 + " -1"
        card = Input([in_str], BlockType.DATA)
        transform = Transform(card)
        output = transform.format_for_mcnp_input((6, 2, 0))
        answers = [in_str.replace("5", "2")]
        self.assertEqual(output[0], in_str)
        transform.number = 2
        output = transform.format_for_mcnp_input((6, 2, 0))
        self.assertEqual(len(output), len(answers))
        for i, line in enumerate(output):
            self.assertEqual(line, answers[i])
        in_str = "*tr5 " + "0.0 " * 3 + "0.0 " * 9 + " -1"
        answers = [in_str.replace("5", "2")]
        card = Input([in_str], BlockType.DATA)
        transform = Transform(card)
        transform.number = 2
        output = transform.format_for_mcnp_input((6, 2, 0))
        self.assertEqual(output, answers)

    def test_transform_equivalent(self):
        in_str = "tr5 " + "0.0 " * 3 + "0.0 " * 9 + " -1"
        card = Input([in_str], BlockType.DATA)
        base = Transform(card)
        self.assertTrue(base.equivalent(base, 1e-6))
        # test different degrees
        in_str = "*tr5 " + "0.0 " * 3 + "0.0 " * 9 + " -1"
        card = Input([in_str], BlockType.DATA)
        compare = Transform(card)
        self.assertFalse(base.equivalent(compare, 1e-3))
        # test different main_aux
        in_str = "tr5 " + "0.0 " * 3 + "0.0 " * 9
        card = Input([in_str], BlockType.DATA)
        compare = Transform(card)
        self.assertFalse(base.equivalent(compare, 1e-3))
        # test different displace
        in_str = "tr5 " + "1.0 " * 3 + "0.0 " * 9 + "-1"
        card = Input([in_str], BlockType.DATA)
        compare = Transform(card)
        self.assertFalse(base.equivalent(compare, 1e-3))
        # test different rotation
        in_str = "tr5 " + "0.0 " * 3 + "1.0 " * 9 + "-1"
        card = Input([in_str], BlockType.DATA)
        compare = Transform(card)
        self.assertFalse(base.equivalent(compare, 1e-3))
        # test different rotation
        in_str = "tr5 " + "0.0 " * 3
        card = Input([in_str], BlockType.DATA)
        compare = Transform(card)
        compare.is_main_to_aux = False
        self.assertFalse(base.equivalent(compare, 1e-3))

    def test_transform_update_values(self):
        in_str = "tr5 " + "0.0 " * 3 + "0.0 " * 9 + " -1"
        card = Input([in_str], BlockType.DATA)
        base = Transform(card)
        base._update_values()
        self.assertEqual(base._tree["classifier"].modifier.value, "")
        self.assertEqual(len(base.data), 13)
        # test with no rotation matrix start
        in_str = "tr5 " + "0.0 " * 3
        card = Input([in_str], BlockType.DATA)
        base = Transform(card)
        test = copy.deepcopy(base)
        test.rotation_matrix = np.ones(9)
        test._update_values()
        self.assertEqual(len(test.data), 12)
        self.assertEqual(test.data[-1].value, 1.0)
        test.is_main_to_aux = False
        test._update_values()
        self.assertEqual(len(test.data), 13)
        self.assertTrue(test.data[-1].is_negative)
        # test partial rotation matrix start
        in_str = "tr5 " + "0.0 " * 3 + "0.0 " * 5
        card = Input([in_str], BlockType.DATA)
        base = Transform(card)
        base.rotation_matrix = np.ones(9)
        base._update_values()
        self.assertEqual(len(base.data), 12)
        self.assertEqual(base.data[-1].value, 1.0)



# Copyright 2024, Battelle Energy Alliance, LLC All Rights Reserved.
import copy
import montepy
import montepy.cells
from montepy.errors import NumberConflictError
import unittest


class TestNumberedObjectCollection(unittest.TestCase):
    def setUp(self):
        self.simple_problem = montepy.read_input("tests/inputs/test.imcnp")

    def test_bad_init(self):
        with self.assertRaises(TypeError):
            montepy.cells.Cells(5)

    def test_numbers(self):
        cell_numbers = [1, 2, 3, 99, 5]
        surf_numbers = [1000, 1005, 1010]
        mat_numbers = [1, 2, 3]
        problem = self.simple_problem
        self.assertEqual(list(problem.cells.numbers), cell_numbers)
        self.assertEqual(list(problem.surfaces.numbers), surf_numbers)
        self.assertEqual(list(problem.materials.numbers), mat_numbers)

    def test_number_conflict_init(self):
        cells = list(self.simple_problem.cells)
        cells.append(cells[1])
        with self.assertRaises(NumberConflictError):
            montepy.cells.Cells(cells)

    def test_check_number(self):
        with self.assertRaises(NumberConflictError):
            self.simple_problem.cells.check_number(1)
        with self.assertRaises(TypeError):
            self.simple_problem.cells.check_number("5")
        # testing a number that shouldn't conflict to ensure error isn't raised
        self.simple_problem.cells.check_number(20)

    def test_objects(self):
        generated = list(self.simple_problem.cells)
        objects = self.simple_problem.cells.objects
        self.assertEqual(generated, objects)

    def test_pop(self):
        cells = copy.deepcopy(self.simple_problem.cells)
        size = len(cells)
        target = list(cells)[-1]
        popped = cells.pop()
        self.assertEqual(target, popped)
        self.assertEqual(size - 1, len(cells))
        with self.assertRaises(TypeError):
            cells.pop("hi")

    def test_extend(self):
        surfaces = copy.deepcopy(self.simple_problem.surfaces)
        extender = list(surfaces)[0:2]
        size = len(surfaces)
        with self.assertRaises(NumberConflictError):
            surfaces.extend(extender)
        self.assertEqual(len(surfaces), size)
        extender = copy.deepcopy(extender)
        extender[0].number = 50
        extender[1].number = 60
        surfaces.extend(extender)
        self.assertEqual(len(surfaces), size + 2)
        # force a num_cache miss
        extender = copy.deepcopy(extender)
        for surf in extender:
            surf._problem = None
        surfaces[1000].number = 1
        extender[0].number = 1000
        extender[1].number = 70
        surfaces.extend(extender)
        self.assertEqual(len(surfaces), size + 4)
        with self.assertRaises(TypeError):
            surfaces.extend(5)
        with self.assertRaises(TypeError):
            surfaces.extend([5])

    def test_iter(self):
        size = len(self.simple_problem.cells)
        counter = 0
        for cell in self.simple_problem.cells:
            counter += 1
        self.assertEqual(size, counter)

    def test_append(self):
        cells = copy.deepcopy(self.simple_problem.cells)
        cell = copy.deepcopy(cells[1])
        size = len(cells)
        with self.assertRaises(NumberConflictError):
            cells.append(cell)
        with self.assertRaises(TypeError):
            cells.append(5)
        cell.number = 20
        cells.append(cell)
        self.assertEqual(len(cells), size + 1)

    def test_append_renumber(self):
        cells = copy.deepcopy(self.simple_problem.cells)
        size = len(cells)
        cell = copy.deepcopy(cells[1])
        cell.number = 20
        cells.append_renumber(cell)
        self.assertEqual(len(cells), size + 1)
        with self.assertRaises(TypeError):
            cells.append_renumber(5)
        with self.assertRaises(TypeError):
            cells.append_renumber(cell, "hi")
        cell = copy.deepcopy(cell)
        cell._problem = None
        cell.number = 1
        cells.append_renumber(cell)
        self.assertEqual(cell.number, 4)
        self.assertEqual(len(cells), size + 2)

    def test_request_number(self):
        cells = self.simple_problem.cells
        self.assertEqual(cells.request_number(6), 6)
        self.assertEqual(cells.request_number(1), 4)
        self.assertEqual(cells.request_number(99, 6), 105)
        with self.assertRaises(TypeError):
            cells.request_number("5")
        with self.assertRaises(TypeError):
            cells.request_number(1, "5")

    def test_next_number(self):
        cells = self.simple_problem.cells
        self.assertEqual(cells.next_number(), 100)
        self.assertEqual(cells.next_number(6), 105)
        with self.assertRaises(TypeError):
            cells.next_number("5")
        with self.assertRaises(ValueError):
            cells.next_number(-1)

    def test_getitem(self):
        cells = self.simple_problem.cells
        list_version = list(cells)
        self.assertEqual(cells[1], list_version[0])
        # force stale cache misses
        cells[1].number = 20
        with self.assertRaises(KeyError):
            cells[1]
        # force cache miss
        self.assertEqual(cells[20], list_version[0])
        with self.assertRaises(TypeError):
            cells["5"]

    def test_delete(self):
        cells = copy.deepcopy(self.simple_problem.cells)
        size = len(cells)
        del cells[1]
        self.assertEqual(size - 1, len(cells))
        with self.assertRaises(TypeError):
            del cells["5"]

    def test_setitem(self):
        cells = copy.deepcopy(self.simple_problem.cells)
        cell = cells[1]
        size = len(cells)
        with self.assertRaises(NumberConflictError):
            cells[1] = cell
        with self.assertRaises(TypeError):
            cells[1] = 5
        with self.assertRaises(TypeError):
            cells["1"] = cell
        cell = copy.deepcopy(cell)
        cell.number = 20
        cells[50] = cell
        self.assertEqual(len(cells), size + 1)

    def test_iadd(self):
        cells = copy.deepcopy(self.simple_problem.cells)
        list_cells = list(cells)
        size = len(cells)
        with self.assertRaises(NumberConflictError):
            cells += list_cells
        with self.assertRaises(NumberConflictError):
            cells += montepy.cells.Cells(list_cells)

        with self.assertRaises(TypeError):
            cells += 5
        with self.assertRaises(TypeError):
            cells += [5]

        list_cells = [copy.deepcopy(cells[1])]
        list_cells[0].number = 20
        cells += list_cells
        self.assertEqual(len(cells), size + 1)

        this_problem = copy.deepcopy(self.simple_problem)
        for cell in this_problem.cells:
            cell.number += 1000
        this_problem.cells += self.simple_problem.cells
        self.assertEqual(len(this_problem.cells), size * 2)

    def test_slice(self):
        test_numbers = [c.number for c in self.simple_problem.cells[1:5]]
        self.assertEqual([1, 2, 3, 5], test_numbers)
        test_numbers = [c.number for c in self.simple_problem.cells[2:]]
        self.assertEqual([2, 3, 5, 99], test_numbers)
        test_numbers = [c.number for c in self.simple_problem.cells[::-3]]
        self.assertEqual([99, 3], test_numbers)
        test_numbers = [c.number for c in self.simple_problem.cells[:6:3]]
        self.assertEqual([3], test_numbers)
        test_numbers = [c.number for c in self.simple_problem.cells[5::-1]]
        self.assertEqual([5, 3, 2, 1], test_numbers)
        test_numbers = [s.number for s in self.simple_problem.surfaces[1000::10]]
        self.assertEqual([1000, 1010], test_numbers)
        test_numbers = [s.number for s in self.simple_problem.surfaces[:]]
        self.assertEqual([1000, 1005, 1010], test_numbers)
        test_numbers = [m.number for m in self.simple_problem.materials[:2]]
        self.assertEqual([1, 2], test_numbers)
        test_numbers = [m.number for m in self.simple_problem.materials[::2]]
        self.assertEqual([2], test_numbers)

    def test_get(self):
        cell_found = self.simple_problem.cells.get(1)
        self.assertEqual(self.simple_problem.cells[1], cell_found)
        surf_not_found = self.simple_problem.surfaces.get(39)  # 39 buried, 0 found
        self.assertIsNone(surf_not_found)
        default_mat = self.simple_problem.materials[3]
        self.assertEqual(
            self.simple_problem.materials.get(42, default_mat), default_mat
        )

    def test_keys(self):
        cell_nums = []
        for c in self.simple_problem.cells:
            cell_nums.append(c.number)
        cell_keys = []
        for k in self.simple_problem.cells.keys():
            cell_keys.append(k)
        self.assertEqual(cell_nums, cell_keys)

    def test_values(self):
        list_cells = list(self.simple_problem.cells)
        list_values = list(self.simple_problem.cells.values())
        self.assertEqual(list_cells, list_values)

    def test_items(self):
        zipped = zip(
            self.simple_problem.cells.keys(), self.simple_problem.cells.values()
        )
        cell_items = self.simple_problem.cells.items()
        self.assertTupleEqual(tuple(zipped), tuple(cell_items))

    def test_surface_generators(self):
        answer_num = [1000, 1010]
        spheres = list(self.simple_problem.surfaces.so)
        self.assertEqual(len(answer_num), len(spheres))
        for i, sphere in enumerate(spheres):
            self.assertEqual(answer_num[i], sphere.number)

    def test_number_adding_concurancy(self):
        surfaces = copy.deepcopy(self.simple_problem.surfaces)
        new_surf = copy.deepcopy(surfaces[1005])
        new_surf.number = 5
        surfaces.append(new_surf)
        size = len(surfaces)
        new_surf = copy.deepcopy(new_surf)
        with self.assertRaises(NumberConflictError):
            surfaces.append(new_surf)
        surfaces.append_renumber(new_surf)
        self.assertEqual(len(surfaces), size + 1)
        self.assertEqual(new_surf.number, 6)

    def test_str(self):
        cells = self.simple_problem.cells
        self.assertEqual(str(cells), "Cells: [1, 2, 3, 99, 5]")
        key_phrases = [
            "Numbered_object_collection: obj_class: <class 'montepy.cell.Cell'>",
            "Objects: [CELL: 1",
            "Number cache: {1: CELL: 1",
        ]
        for phrase in key_phrases:
            self.assertIn(phrase, repr(cells))



# Copyright 2024, Battelle Energy Alliance, LLC All Rights Reserved.
from abc import ABC, abstractmethod
import typing
import montepy
from montepy.numbered_mcnp_object import Numbered_MCNP_Object
from montepy.errors import *


class NumberedObjectCollection(ABC):
    """A collections of MCNP objects.

    It quacks like a dict, it acts like a dict, but it's a list.

    The items in the collection are accessible by their number.
    For instance to get the Cell with a number of 2 you can just say:

    ``problem.cells[2]``

    You can also add delete items like you would in a dictionary normally.

    Unlike dictionaries this collection also supports slices e.g., ``[1:3]``.
    This will return a new :class:`NumberedObjectCollection` with objects
    that have cell numbers that fit that slice. If a number is in a slice that
    is not an actual object it will just be skipped.

    Because MCNP numbered objects start at 1, so do the indices.
    The slices are effectively 1-based and endpoint-inclusive.
    This means rather than the normal behavior of [0:5] excluding the index
    5, 5 would be included.

    :param obj_class: the class of numbered objects being collected
    :type obj_class: type
    :param objects: the list of cells to start with if needed
    :type objects: list
    :param problem: the problem to link this collection to.
    :type problem: MCNP_Problem
    """

    def __init__(self, obj_class, objects=None, problem=None):
        self.__num_cache = {}
        assert issubclass(obj_class, Numbered_MCNP_Object)
        self._obj_class = obj_class
        self._objects = []
        self._problem = problem
        if objects:
            if not isinstance(objects, list):
                raise TypeError("NumberedObjectCollection must be built from a list")
            for obj in objects:
                if not isinstance(obj, obj_class):
                    raise TypeError(
                        f"The object: {obj} being added to a NumberedObjectCollection is not of type {obj_class}"
                    )
                if obj.number in self.__num_cache:
                    raise NumberConflictError(
                        (
                            f"When building {self} there was a numbering conflict between: "
                            f"{obj} and {self[obj.number]}"
                        )
                    )
                self.__num_cache[obj.number] = obj
            self._objects = objects

    def link_to_problem(self, problem):
        """Links the card to the parent problem for this card.

        This is done so that cards can find links to other objects.

        :param problem: The problem to link this card to.
        :type problem: MCNP_Problem
        """
        if not isinstance(problem, montepy.mcnp_problem.MCNP_Problem):
            raise TypeError("problem must be an MCNP_Problem")
        self._problem = problem

    @property
    def numbers(self):
        """
        A generator of the numbers being used.

        :rtype: generator
        """
        self.__num_cache
        for obj in self._objects:
            # update cache every time we go through all objects
            self.__num_cache[obj.number] = obj
            yield obj.number

    def check_number(self, number):
        """Checks if the number is already in use, and if so raises an error.

        :param number: The number to check.
        :type number: int
        :raises NumberConflictError: if this number is in use.
        """
        if not isinstance(number, int):
            raise TypeError("The number must be an int")
        if number in self.numbers:
            raise NumberConflictError(
                f"Number {number} is already in use for the collection: {type(self)} by {self[number]}"
            )

    @property
    def objects(self):
        """
        Returns a shallow copy of the internal objects list.

        The list object is a new instance, but the underlying objects
        are the same.

        :rtype: list
        """
        return self._objects[:]

    def pop(self, pos=-1):
        """
        Pop the final items off of the collection

        :param pos: The index of the element to pop from the internal list.
        :type pos: int
        :return: the final elements
        :rtype: Numbered_MCNP_Object
        """
        if not isinstance(pos, int):
            raise TypeError("The index for popping must be an int")
        obj = self._objects.pop(pos)
        self.__num_cache.pop(obj.number, None)
        return obj

    def clear(self):
        """
        Removes all objects from this collection.
        """
        self._objects.clear()
        self.__num_cache.clear()

    def extend(self, other_list):
        """
        Extends this collection with another list.

        :param other_list: the list of objects to add.
        :type other_list: list
        :raises NumberConflictError: if these items conflict with existing elements.
        """
        if not isinstance(other_list, (list, type(self))):
            raise TypeError("The extending list must be a list")
        for obj in other_list:
            if not isinstance(obj, self._obj_class):
                raise TypeError(
                    "The object in the list {obj} is not of type: {self._obj_class}"
                )
            if obj.number in self.numbers:
                raise NumberConflictError(
                    (
                        f"When adding to {type(self)} there was a number collision due to "
                        f"adding {obj} which conflicts with {self[obj.number]}"
                    )
                )
            # if this number is a ghost; remove it.
            else:
                self.__num_cache.pop(obj.number, None)
        self._objects.extend(other_list)
        if self._problem:
            for obj in other_list:
                obj.link_to_problem(self._problem)

    def remove(self, delete):
        """
        Removes the given object from the collection.

        :param delete: the object to delete
        :type delete: Numbered_MCNP_Object
        """
        self.__num_cache.pop(delete.number, None)
        self._objects.remove(delete)

    def __iter__(self):
        self._iter = self._objects.__iter__()
        return self._iter

    def __str__(self):
        base_class_name = self.__class__.__name__
        numbers = list(self.numbers)
        return f"{base_class_name}: {numbers}"

    def __repr__(self):
        return (
            f"Numbered_object_collection: obj_class: {self._obj_class}, problem: {self._problem}\n"
            f"Objects: {self._objects}\n"
            f"Number cache: {self.__num_cache}"
        )

    def append(self, obj):
        """Appends the given object to the end of this collection.

        :param obj: the object to add.
        :type obj: Numbered_MCNP_Object
        :raises NumberConflictError: if this object has a number that is already in use.
        """
        if not isinstance(obj, self._obj_class):
            raise TypeError(f"object being appended must be of type: {self._obj_class}")
        if obj.number in self.numbers:
            raise NumberConflictError(
                (
                    "There was a numbering conflict when attempting to add "
                    f"{obj} to {type(self)}. Conflict was with {self[obj.number]}"
                )
            )
        else:
            self.__num_cache[obj.number] = obj
        self._objects.append(obj)
        if self._problem:
            obj.link_to_problem(self._problem)

    def append_renumber(self, obj, step=1):
        """Appends the object, but will renumber the object if collision occurs.

        This behaves like append, except if there is a number collision the object will
        be renumbered to an available number. The number will be incremented by step
        until an available number is found.

        :param obj: The MCNP object being added to the collection.
        :type obj: Numbered_MCNP_Object
        :param step: the incrementing step to use to find a new number.
        :type step: int
        :return: the number for the object.
        :rtype: int
        """
        if not isinstance(obj, self._obj_class):
            raise TypeError(f"object being appended must be of type: {self._obj_class}")
        if not isinstance(step, int):
            raise TypeError("The step number must be an int")
        number = obj.number
        try:
            self.append(obj)
        except NumberConflictError:
            number = self.request_number(number, step)
            obj.number = number
            self.append(obj)

        if self._problem:
            obj.link_to_problem(self._problem)
        return number

    def request_number(self, start_num=1, step=1):
        """Requests a new available number.

        This method does not "reserve" this number. Objects
        should be immediately added to avoid possible collisions
        caused by shifting numbers of other objects in the collection.

        :param start_num: the starting number to check.
        :type start_num: int
        :param step: the increment to jump by to find new numbers.
        :type step: int
        :returns: an available number
        :rtype: int
        """
        if not isinstance(start_num, int):
            raise TypeError("start_num must be an int")
        if not isinstance(step, int):
            raise TypeError("step must be an int")
        number = start_num
        while number in self.numbers:
            number += step
        return number

    def next_number(self, step=1):
        """Get the next available number, based on the maximum number.

        This works by finding the current maximum number, and then adding the
        stepsize to it.

        :param step: how much to increase the last number by
        :type step: int
        """
        if not isinstance(step, int):
            raise TypeError("step must be an int")
        if step <= 0:
            raise ValueError("step must be > 0")
        return max(self.numbers) + step

    def __get_slice(self, i: slice):
        """Get a new NumberedObjectCollection over a slice of numbers

        This method implements usage like:
            >>> NumberedObjectCollection[1:49:12]
            [1, 13, 25, 37, 49]

        Any ``slice`` object may be passed.
        The indices are the object numbers.
        Because MCNP numbered objects start at 1, so do the indices.
        They are effectively 1-based and endpoint-inclusive.

        :rtype: NumberedObjectCollection
        """
        rstep = i.step if i.step is not None else 1
        rstart = i.start
        rstop = i.stop
        if rstep < 0:  # Backwards
            if rstart is None:
                rstart = max(self.numbers)
            if rstop is None:
                rstop = min(self.numbers)
            rstop -= 1
        else:  # Forwards
            if rstart is None:
                rstart = 0
            if rstop is None:
                rstop = max(self.numbers)
            rstop += 1
        numbered_objects = []
        for num in range(rstart, rstop, rstep):
            obj = self.get(num)
            if obj is not None:
                numbered_objects.append(obj)
        # obj_class is always implemented in child classes.
        return type(self)(numbered_objects)

    def __getitem__(self, i):
        if isinstance(i, slice):
            return self.__get_slice(i)
        elif not isinstance(i, int):
            raise TypeError("index must be an int or slice")
        ret = self.get(i)
        if ret is None:
            raise KeyError(f"Object with number {i} not found in {type(self)}")
        return ret

    def __delitem__(self, idx):
        if not isinstance(idx, int):
            raise TypeError("index must be an int")
        obj = self[idx]
        self.__num_cache.pop(obj.number, None)
        idx = self._objects.index(obj)
        del self._objects[idx]

    def __setitem__(self, key, newvalue):
        if not isinstance(key, int):
            raise TypeError("index must be an int")
        self.append(newvalue)

    def __len__(self):
        return len(self._objects)

    def __iadd__(self, other):
        if not isinstance(other, (type(self), list)):
            raise TypeError(f"Appended item must be a list or of type {type(self)}")
        for obj in other:
            if not isinstance(obj, self._obj_class):
                raise TypeError(
                    f"Appended object {obj} must be of type: {self._obj_class}"
                )
        if isinstance(other, type(self)):
            other_list = other.objects
        else:
            other_list = other
        for obj in other_list:
            if obj.number in self.numbers:
                raise NumberConflictError(
                    (
                        "There was a numbering conflict when attempting to add "
                        f"{obj} to {type(self)}. Conflict was with {self[obj.number]}"
                    )
                )
            else:
                self.__num_cache[obj.number] = obj
        self._objects += other_list
        if self._problem:
            for obj in other_list:
                obj.link_to_problem(self._problem)
        return self

    def __contains__(self, other):
        return other in self._objects

    def get(self, i: int, default=None) -> (Numbered_MCNP_Object, None):
        """
        Get ``i`` if possible, or else return ``default``.

        :param i: number of the object to get, not it's location in the internal list
        :type i: int
        :param default: value to return if not found
        :type default: object

        :rtype: Numbered_MCNP_Object
        """
        try:
            ret = self.__num_cache[i]
            if ret.number == i:
                return ret
        except KeyError:
            pass
        for obj in self._objects:
            if obj.number == i:
                self.__num_cache[i] = obj
                return obj
        return default

    def keys(self) -> typing.Generator[int, None, None]:
        """
        Get iterator of the collection's numbers.

        :rtype: int
        """
        for o in self._objects:
            yield o.number

    def values(self) -> typing.Generator[Numbered_MCNP_Object, None, None]:
        """
        Get iterator of the collection's objects.

        :rtype: Numbered_MCNP_Object
        """
        for o in self._objects:
            yield o

    def items(
        self,
    ) -> typing.Generator[typing.Tuple[int, Numbered_MCNP_Object], None, None]:
        """
        Get iterator of the collections (number, object) pairs.

        :rtype: tuple(int, MCNP_Object)
        """
        for o in self._objects:
            yield o.number, o



# Copyright 2024, Battelle Energy Alliance, LLC All Rights Reserved.
from montepy.input_parser.parser_base import MCNP_Parser
from montepy.input_parser.tokens import SurfaceLexer
from montepy.input_parser import syntax_node


class SurfaceParser(MCNP_Parser):
    """
    A parser for MCNP surfaces.

    .. versionadded:: 0.2.0
        This was added with the major parser rework.

    :rtype: SyntaxNode
    """

    debugfile = None

    @_(
        "surface_id SURFACE_TYPE padding number_sequence",
        "padding surface_id SURFACE_TYPE padding number_sequence",
        "surface_id number_phrase SURFACE_TYPE padding number_sequence",
        "padding surface_id number_phrase SURFACE_TYPE padding number_sequence",
    )
    def surface(self, p):
        ret = {}
        if isinstance(p[0], syntax_node.PaddingNode):
            ret["start_pad"] = p[0]
        else:
            ret["start_pad"] = syntax_node.PaddingNode()
        ret["surface_num"] = p.surface_id
        if hasattr(p, "number_phrase"):
            ret["pointer"] = p.number_phrase
        else:
            ret["pointer"] = syntax_node.ValueNode(None, int)
        if hasattr(p, "padding"):
            padding = p.padding
        else:
            padding = p.padding1
        ret["surface_type"] = syntax_node.ValueNode(p.SURFACE_TYPE, str, padding)
        ret["data"] = p.number_sequence
        return syntax_node.SyntaxNode("surface", ret)

    @_('"*" number_phrase', '"+" number_phrase', "number_phrase")
    def surface_id(self, p):
        ret = {}
        if isinstance(p[0], str) and p[0] in {"*", "+"}:
            ret["modifier"] = syntax_node.ValueNode(p[0], str)
        else:
            ret["modifier"] = syntax_node.ValueNode(None, str)

        ret["number"] = p.number_phrase
        return syntax_node.SyntaxNode("surface_number", ret)



# Copyright 2024, Battelle Energy Alliance, LLC All Rights Reserved.
from unittest import TestCase

from montepy.utilities import fortran_float


class testFortranFloat(TestCase):
    def test_normal_float_parse(self):
        tests = {"123": 123, "1.23": 1.23, "1.2e+3": 1.2e3, "1.2e-3": 1.2e-3}
        for test_string in tests:
            self.assertAlmostEqual(fortran_float(test_string), tests[test_string])

    def test_stupid_float_parse(self):
        tests = {"1.2+3": 1.2e3, "1.2-3": 1.2e-3, "-2-3": -2.0e-3}
        for test_string in tests:
            self.assertAlmostEqual(fortran_float(test_string), tests[test_string])

    def test_raise_error(self):
        with self.assertRaises(ValueError):
            fortran_float("Dog")



# Copyright 2024, Battelle Energy Alliance, LLC All Rights Reserved.
from unittest import TestCase

import montepy

from montepy.mcnp_problem import MCNP_Problem


class testMCNP_problem(TestCase):
    def test_problem_init(self):
        problem = MCNP_Problem("tests/inputs/test.imcnp")
        self.assertIsInstance(
            problem.input_file, montepy.input_parser.input_file.MCNP_InputFile
        )
        self.assertEqual(problem.input_file.path, "tests/inputs/test.imcnp")
        self.assertEqual(problem.input_file.name, "tests/inputs/test.imcnp")
        self.assertEqual(problem.mcnp_version, (6, 2, 0))

    def test_problem_str(self):
        file_name = "tests/inputs/test.imcnp"
        problem = MCNP_Problem(file_name)
        self.assertTrue(f"MCNP problem for: {file_name}" in str(problem))
